import functools


def parameter_annotations(**annotations):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        if annotations:
            wrapper.parameter_annotations = annotations
        return wrapper

    return decorator


def parameter_constraints(**constraints):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        # Attach only if constraints are provided
        if constraints:
            wrapper.parameter_constraints = constraints
        return wrapper

    return decorator



import os
from pathlib import Path
from typing import Callable, Dict

from PyQt6.QtWidgets import QVBoxLayout

from sonaris.frontend.managers.device import DeviceManager
from sonaris.frontend.widgets.set_mock import MockHardwareWidget
from sonaris.frontend.widgets.set_state import SettingsStateWidget
from sonaris.frontend.widgets.templates import BasePage
from sonaris.defaults import SETTINGS_FILE

class SettingsPage(BasePage):
    PAGE_NAME = "Settings"

    def __init__(
        self,
        device_managers: Dict[str, DeviceManager] = None,
        parent=None,
        args_dict: dict = None,
        root_callback: Callable = None,
    ):
        super().__init__(
            parent=parent, args_dict=args_dict, root_callback=root_callback
        )
        self.device_managers = device_managers
        self.settings_widget = SettingsStateWidget(
            settings_file=SETTINGS_FILE
        )
        self.mock_hardware_widget = None  # Will be initialized if hardware_mock is True
        self.initUI()
        self.setLayout(self.main_layout)

    def initUI(self):
        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.settings_widget)

        # Add the mock hardware settings widget if hardware_mock is True
        if self.args_dict and self.args_dict.get("hardware_mock"):
            self.mock_hardware_widget = MockHardwareWidget(self.device_managers)
            self.main_layout.addWidget(self.mock_hardware_widget)

    def update(self):
        pass  # Implement any update logic here if needed



from datetime import datetime

import pyqtgraph as pg
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QComboBox,
    QDoubleSpinBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from sonaris.device.dg4202 import DG4202
from sonaris.defaults import DECIMAL_POINTS, NOT_FOUND_STRING
from sonaris.frontend.managers.dg4202 import DG4202Manager
from sonaris.frontend.pages import plotter
from sonaris.utils.log import get_logger

logger = get_logger()


class DG4202SignalGeneratorWidget(QWidget):
    def __init__(self, dg4202_manager: DG4202Manager, parent=None):
        super().__init__(parent=parent)
        self.sweep_plot_data = {1: None, 2: None}
        self.waveform_plot_data = {1: None, 2: None}
        self.channel_count = 2
        self.link_channel = False
        self.dg4202_manager = dg4202_manager
        self.all_parameters = self.dg4202_manager.get_data()
        self.input_objects = {1: {}, 2: {}}
        self.initUI()

    def check_connection(self) -> bool:
        self.dg4202_manager.device = self.dg4202_manager.get_device()
        self.all_parameters = self.dg4202_manager.get_data()

        if self.dg4202_manager.device is None:
            return False

        is_alive = self.dg4202_manager.device.is_connection_alive()
        if not is_alive:
            self.dg4202_manager.device = None
        return is_alive

    def create_widgets(self):
        # A dictionary to store widgets for each channel by channel number
        self.controls = {}

        for channel in range(1, self.channel_count + 1):
            # Main controls widget
            main_controls_widget = self.generate_main_controls(channel)

            # Individual control widgets
            waveform_control = self.generate_waveform_control(channel)
            sweep_control = self.generate_sweep_control(channel)

            # Store widgets in dictionary
            self.controls[channel] = {
                "main_controls": main_controls_widget,
                "waveform_control": waveform_control,
                "sweep_control": sweep_control,
            }

    def initUI(self):
        self.create_widgets()
        self.main_layout = QVBoxLayout()
        self.check_connection()
        self.status_label = QLabel("")
        self.main_layout.addWidget(self.status_label)
        for _, widgets in self.controls.items():
            # Add main controls for the channel

            tab_widget = QTabWidget()
            tab_widget.addTab(widgets["waveform_control"], "Default")
            tab_widget.addTab(widgets["sweep_control"], "Sweep")
            tab_widget.setMaximumHeight(250)
            self.main_layout.addWidget(tab_widget)
            tab_widget.currentChanged.connect(self.on_tab_changed)
            self.main_layout.addWidget(widgets["main_controls"])

        self.main_layout.setSpacing(5)  # Reduced spacing
        self.main_layout.setContentsMargins(5, 5, 5, 5)  # Reduced margins

        self.setLayout(self.main_layout)

    def generate_main_controls(self, channel: int) -> QWidget:

        # Setting up main layout
        main_layout = QVBoxLayout()

        control_layout = QHBoxLayout()
        # Adding tabs for modes
        # Add more tabs as needed
        self.input_objects[channel]["TOGGLE_OUTPUT"] = QPushButton(
            f"Output CH{channel}"
        )
        self.input_objects[channel]["TOGGLE_OUTPUT"].clicked.connect(
            lambda: self.toggle_output(channel)
        )
        self.update_button_state(channel)
        control_layout.addWidget(self.input_objects[channel]["TOGGLE_OUTPUT"])

        main_layout.addLayout(control_layout)

        # Create a QWidget to return
        main_controls_widget = QWidget()
        main_controls_widget.setLayout(main_layout)
        return main_controls_widget

    def generate_sweep_control(self, channel: int) -> QWidget:
        """
        Generate the control components for the sweep mode.

        Parameters:
            channel (int): The channel number.

        Returns:
            QWidget: The widget containing the sweep control components.
        """

        # ---- LEFT COLUMN ---- #
        left_first_column_layout = QFormLayout()
        left_second_column_layout = QFormLayout()

        # Sweep input
        duration_input = QLineEdit(self)
        duration_input.setPlaceholderText(f"Sweep duration frequency CH{channel}")
        self.input_objects[channel]["TIME"] = duration_input
        # Set the text of the sweep input to the current value
        current_sweep_value = str(
            self.all_parameters.get(f"{channel}", {})
            .get("mode", {})
            .get("parameters", {})
            .get("sweep", {})
            .get("TIME", 0.0)
        )
        duration_input.setText(current_sweep_value)
        left_first_column_layout.addRow("Sweep (s):", duration_input)

        # Return time input
        rtime_input = QLineEdit(self)
        rtime_input.setPlaceholderText(f"Sweep return time CH{channel}")
        self.input_objects[channel]["RTIME"] = rtime_input
        # Set the text of the return time input to the current value
        current_rtime_value = str(
            self.all_parameters.get(f"{channel}", {})
            .get("mode", {})
            .get("parameters", {})
            .get("sweep", {})
            .get("RTIME", 0.0)
        )
        rtime_input.setText(current_rtime_value)
        left_first_column_layout.addRow("Return (ms):", rtime_input)

        # Start frequency input
        fstart_input = QLineEdit(self)
        fstart_input.setPlaceholderText(f"Sweep start frequency CH{channel}")
        self.input_objects[channel]["FSTART"] = fstart_input
        # Set the text of the start frequency input to the current value
        current_fstart_value = str(
            self.all_parameters.get(f"{channel}", {})
            .get("mode", {})
            .get("parameters", {})
            .get("sweep", {})
            .get("FSTART", 0.0)
        )
        fstart_input.setText(current_fstart_value)
        left_first_column_layout.addRow("Start (Hz):", fstart_input)

        # Stop frequency input
        fstop_input = QLineEdit(self)
        fstop_input.setPlaceholderText(f"Sweep stop frequency CH{channel}")
        self.input_objects[channel]["FSTOP"] = fstop_input
        # Set the text of the stop frequency input to the current value
        current_fstop_value = str(
            self.all_parameters.get(f"{channel}", {})
            .get("mode", {})
            .get("parameters", {})
            .get("sweep", {})
            .get("FSTOP", 0.0)
        )
        fstop_input.setText(current_fstop_value)
        left_second_column_layout.addRow("Stop (Hz):", fstop_input)

        # Hold Start input
        htime_start_input = QLineEdit(self)
        htime_start_input.setPlaceholderText(f"Start hold time CH{channel}")
        self.input_objects[channel]["HTIME_START"] = htime_start_input
        # Set the text of the start hold time input to the current value
        current_htime_start_value = str(
            self.all_parameters.get(f"{channel}", {})
            .get("mode", {})
            .get("parameters", {})
            .get("sweep", {})
            .get("HTIME_START", 0.0)
        )
        htime_start_input.setText(current_htime_start_value)
        left_second_column_layout.addRow("Start Hold (ms):", htime_start_input)

        # Hold Stop input
        htime_stop_input = QLineEdit(self)
        htime_stop_input.setPlaceholderText(f"Stop hold time CH{channel}")
        self.input_objects[channel]["HTIME_STOP"] = htime_stop_input
        # Set the text of the stop hold time input to the current value
        current_htime_stop_value = str(
            self.all_parameters.get(f"{channel}", {})
            .get("mode", {})
            .get("parameters", {})
            .get("sweep", {})
            .get("HTIME_STOP", 0.0)
        )
        htime_stop_input.setText(current_htime_stop_value)
        left_second_column_layout.addRow("Stop Hold (ms):", htime_stop_input)

        # Set parameters button
        set_parameters_button = QPushButton(f"Set Parameters CH{channel}", self)
        set_parameters_button.clicked.connect(
            lambda: self.on_update_sweep(
                channel,
                duration_input.text(),
                rtime_input.text(),
                fstart_input.text(),
                fstop_input.text(),
                htime_start_input.text(),
                htime_stop_input.text(),
            )
        )
        left_columns_layout = QHBoxLayout()
        left_columns_layout.addLayout(left_first_column_layout)
        left_columns_layout.addLayout(left_second_column_layout)

        # Create a vertical layout to hold the columns and the button
        left_main_layout = QVBoxLayout()
        left_main_layout.addLayout(left_columns_layout)  # Add the columns layout

        # Set parameters button
        left_main_layout.addWidget(
            set_parameters_button
        )  # Adding the button below the columns

        # Wrap the main left layout in a widget
        left_widget = QWidget()
        left_widget.setLayout(left_main_layout)

        # ---- RIGHT COLUMN (Sweep Plot) ---- #
        right_column_layout = QVBoxLayout()
        plot_widget = pg.PlotWidget()
        plot_widget.setLabel("left", "Amplitude", units="V")
        plot_widget.setLabel("bottom", "Time", units="s")
        self.sweep_plot_data[channel] = plot_widget.plot([], pen="y")
        right_column_layout.addWidget(
            plot_widget, alignment=Qt.AlignmentFlag.AlignCenter
        )

        right_widget = QWidget()
        right_widget.setLayout(right_column_layout)

        # ---- COMBINE AND RETURN ---- #
        layout = QHBoxLayout()
        layout.addWidget(left_widget)
        layout.addWidget(right_widget)

        widget = QWidget()
        widget.setLayout(layout)
        self.update_sweep_graph(channel)
        return widget

    def generate_waveform_control(self, channel: int) -> QWidget:
        """
        Generate the control components for the waveform mode.

        Parameters:
            channel (int): The channel number.

        Returns:
            QWidget: The widget containing the waveform control components.
        """

        # Create main container and layout
        channel_widget = QWidget()
        channel_layout = QHBoxLayout()

        # ---- LEFT COLUMN ---- #
        left_column_layout = QFormLayout()

        # Waveform type dropdown
        waveform_type = QComboBox()
        waveform_type.addItems(DG4202.available_waveforms())
        self.input_objects[channel]["WAVEFORM_TYPE"] = waveform_type
        left_column_layout.addRow(f"Waveform Type CH{channel}", waveform_type)

        # Frequency input with QDoubleSpinBox
        freq_input = QDoubleSpinBox()
        freq_input.setDecimals(DECIMAL_POINTS)
        freq_input.setMaximum(self.dg4202_manager.device_type.FREQ_LIMIT)
        freq_input.setValue(
            self.all_parameters.get(f"{channel}", {})
            .get("waveform", {})
            .get("frequency", 0.0)
        )
        self.input_objects[channel]["FREQUENCY"] = freq_input
        left_column_layout.addRow("Frequency (Hz)", freq_input)

        # Amplitude input with QDoubleSpinBox
        amp_input = QDoubleSpinBox()
        amp_input.setDecimals(DECIMAL_POINTS)
        amp_input.setMaximum(5.0)  # 5V limit
        amp_input.setValue(
            self.all_parameters.get(f"{channel}", {})
            .get("waveform", {})
            .get("amplitude", 0.0)
        )
        self.input_objects[channel]["AMPLITUDE"] = amp_input
        left_column_layout.addRow("Amplitude (V)", amp_input)

        # Offset input with QDoubleSpinBox
        offset_input = QDoubleSpinBox()
        offset_input.setDecimals(DECIMAL_POINTS)
        offset_input.setMaximum(5.0)  # 5V limit
        offset_input.setValue(
            self.all_parameters.get(f"{channel}", {})
            .get("waveform", {})
            .get("offset", 0.0)
        )
        self.input_objects[channel]["OFFSET"] = offset_input
        left_column_layout.addRow("Offset (V)", offset_input)

        # Set waveform button
        set_waveform_button = QPushButton(f"Set Waveform CH{channel}")
        set_waveform_button.clicked.connect(
            lambda: self.on_update_waveform(
                channel,
                waveform_type.currentText(),
                float(freq_input.text()),
                float(amp_input.text()),
                float(offset_input.text()),
            )
        )
        left_column_layout.addRow(set_waveform_button)

        # Add to main layout
        left_widget = QWidget()
        left_widget.setLayout(left_column_layout)
        channel_layout.addWidget(left_widget)

        # ---- RIGHT COLUMN (Waveform Plot) ---- #
        right_column_layout = QVBoxLayout()
        plot_widget = pg.PlotWidget()
        plot_widget.setLabel("left", "Amplitude", units="V")
        plot_widget.setLabel("bottom", "Time", units="s")
        self.waveform_plot_data[channel] = plot_widget.plot([], pen="y")
        # Add spacers to constrain the size of the plot widget
        right_column_layout.addSpacerItem(
            QSpacerItem(
                20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
            )
        )
        right_column_layout.addWidget(
            plot_widget, alignment=Qt.AlignmentFlag.AlignCenter
        )
        right_column_layout.addSpacerItem(
            QSpacerItem(
                20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
            )
        )
        # Add to main layout
        right_widget = QWidget()
        right_widget.setLayout(right_column_layout)
        channel_layout.addWidget(right_widget, 0)

        # ---- COMBINE AND RETURN ---- #
        channel_widget.setLayout(channel_layout)
        self.update_waveform_graph(channel)
        return channel_widget

    def on_update_sweep(
        self,
        channel: int,
        sweep: float,
        rtime: float,
        fstart: float,
        fstop: float,
        htime_start: float,
        htime_stop: float,
    ):
        try:
            if self.dg4202_manager.get_device():
                sweep = (
                    float(sweep)
                    if sweep
                    else float(
                        self.all_parameters.get(f"{channel}", {})
                        .get("mode", {})
                        .get("parameters", {})
                        .get("sweep", {})
                        .get("TIME", 0.0)
                    )
                )
                rtime = (
                    float(rtime)
                    if rtime
                    else float(
                        self.all_parameters.get(f"{channel}", {})
                        .get("mode", {})
                        .get("parameters", {})
                        .get("sweep", {})
                        .get("RTIME", 0.0)
                    )
                )
                fstart = (
                    float(fstart)
                    if fstart
                    else float(
                        self.all_parameters.get(f"{channel}", {})
                        .get("mode", {})
                        .get("parameters", {})
                        .get("sweep", {})
                        .get("FSTART", 0.0)
                    )
                )
                fstop = (
                    float(fstop)
                    if fstop
                    else float(
                        self.all_parameters.get(f"{channel}", {})
                        .get("mode", {})
                        .get("parameters", {})
                        .get("sweep", {})
                        .get("FSTOP", 0.0)
                    )
                )
                htime_start = (
                    float(htime_start)
                    if htime_start
                    else float(
                        self.all_parameters.get(f"{channel}", {})
                        .get("mode", {})
                        .get("parameters", {})
                        .get("sweep", {})
                        .get("HTIME_START", 0.0)
                    )
                )
                htime_stop = (
                    float(htime_stop)
                    if htime_stop
                    else float(
                        self.all_parameters.get(f"{channel}", {})
                        .get("mode", {})
                        .get("parameters", {})
                        .get("sweep", {})
                        .get("HTIME_STOP", 0.0)
                    )
                )
                params = {
                    "TIME": sweep,
                    "RTIME": rtime,
                    "FSTART": fstart,
                    "FSTOP": fstop,
                    "HTIME_START": htime_start,
                    "HTIME_STOP": htime_stop,
                }
                logger.info(params)
                self.dg4202_manager.device.set_sweep_parameters(channel, params)
                self.update_sweep_graph(channel)
            else:
                logger.error(f"{NOT_FOUND_STRING} Is device connected?.")
        except Exception as e:
            logger.error(f"Error:{e}")

    def toggle_output(self, channel: int):
        """
        Toggle the output for the given channel.

        Parameters:
            channel (int): The channel number.
        """
        try:
            if self.check_connection():
                set_to = (
                    False
                    if self.all_parameters.get(f"{channel}", {}).get(
                        "output_status", "ERR"
                    )
                    == "ON"
                    else True
                )
                logger.info(
                    f'{channel} is {self.all_parameters.get(f"{channel}", {}).get("output_status", "ERR")} -> {set_to}'
                )
                self.dg4202_manager.device.output_on_off(channel, set_to)
                self.check_connection()  # updates dictionary
                # Update the button state after toggling the output

            else:
                logger.error(f"{NOT_FOUND_STRING} Is device connected?")
            self.update_button_state(channel)
        except Exception as e:
            logger.error(f"Error:{e}")

    def update_button_state(self, channel: int):
        """
        Update the button's appearance and text based on the output status.

        Parameters:
            output_btn (QPushButton): The output button to be updated.
            channel (int): The channel number.
        """
        status = self.all_parameters.get(f"{channel}", {}).get("output_status", "ERR")
        if status == "ON":
            self.input_objects[channel]["TOGGLE_OUTPUT"].setStyleSheet(
                "background-color: white; color: black;"
            )
            self.input_objects[channel]["TOGGLE_OUTPUT"].setText(
                f"Output CH{channel} ON"
            )
        else:
            self.input_objects[channel]["TOGGLE_OUTPUT"].setStyleSheet(
                "background-color: none; color: white;"
            )
            self.input_objects[channel]["TOGGLE_OUTPUT"].setText(
                f"Output CH{channel} OFF"
            )

    def on_tab_changed(self, index):
        # Placeholder for an API call. Replace with actual logic later.
        if index == 0:  # Assuming 0 is the index for the "Sweep" tab
            pass  # API call for "Sweep"
        elif index == 1:  # Assuming 1 is the index for the "Waveform" tab
            pass  # API call for "Waveform"
        # Add more conditions for other tabs

        # Connect the signal

    def on_update_waveform(
        self,
        channel: int,
        waveform_type: str,
        frequency: float,
        amplitude: float,
        offset: float,
    ):
        """
        Update the waveform parameters and plot.

        Parameters:
            channel (int): The channel number.
            waveform_type (str): The selected waveform type.
            frequency (float): The selected frequency.
            amplitude (float): The selected amplitude.
            offset (float): The selected offset.
            self.waveform_plot_data (float): Plot data reference.
        """
        try:
            if self.dg4202_manager.get_device():
                frequency = frequency or float(
                    self.all_parameters[f"{channel}"]["waveform"]["frequency"]
                )
                amplitude = amplitude or float(
                    self.all_parameters[f"{channel}"]["waveform"]["amplitude"]
                )
                offset = offset or float(
                    self.all_parameters[f"{channel}"]["waveform"]["offset"]
                )
                waveform_type = (
                    waveform_type
                    or self.all_parameters[f"{channel}"]["waveform"]["waveform_type"]
                )

                # If a parameter is not set, pass the current value
                self.dg4202_manager.device.set_waveform(
                    channel, waveform_type, frequency, amplitude, offset
                )

                if self.link_channel:
                    self.dg4202_manager.device.set_waveform(
                        2 if channel == 1 else 1,
                        waveform_type,
                        frequency,
                        amplitude,
                        offset,
                    )

                # Update some status label or log if you have one
                status_string = f"[{datetime.now().isoformat()}] Waveform updated."
                # Assuming you have a status_label in your UI
                self.status_label.setText(status_string)
                self.update_waveform_graph(channel)
                logger.info(
                    f"Waveform updated [CH:{channel}] {waveform_type} FREQ:{frequency} AMP:{amplitude} OFF{offset}"
                )
            else:
                self.status_label.setText(NOT_FOUND_STRING)
                logger.error(f"{NOT_FOUND_STRING} Is device connected?")

        except Exception as e:
            logger.error("Error: %s, ", e)

    def update_waveform_graph(self, channel):
        self.check_connection()
        x_data, y_data = plotter.plot_waveform(
            params=self.all_parameters[f"{channel}"]["waveform"]
        )
        self.waveform_plot_data[channel].setData(x_data, y_data)

    def update_sweep_graph(self, channel):
        self.check_connection()
        x_data, y_data = plotter.plot_sweep(
            start_frequency=self.all_parameters[f"{channel}"]["mode"]["parameters"][
                "sweep"
            ]["FSTART"],
            stop_frequency=self.all_parameters[f"{channel}"]["mode"]["parameters"][
                "sweep"
            ]["FSTOP"],
            duration=self.all_parameters[f"{channel}"]["mode"]["parameters"]["sweep"][
                "TIME"
            ],
            rtime=self.all_parameters[f"{channel}"]["mode"]["parameters"]["sweep"][
                "RTIME"
            ],
            htime_start=self.all_parameters[f"{channel}"]["mode"]["parameters"][
                "sweep"
            ]["HTIME_START"],
            htime_stop=self.all_parameters[f"{channel}"]["mode"]["parameters"]["sweep"][
                "HTIME_STOP"
            ],
        )
        self.sweep_plot_data[channel].setData(x_data, y_data)

    def update(self):
        if self.check_connection():
            for channel in range(1, self.channel_count + 1):
                self.update_waveform_graph(channel)
                self.update_sweep_graph(channel)
                self.update_button_state(channel)
                self.update_input_fields(channel)

    def update_input_fields(self, channel: int):
        sweep_parameters = self.all_parameters[f"{channel}"]["mode"]["parameters"][
            "sweep"
        ]
        waveform_parameters = self.all_parameters[f"{channel}"]["waveform"]

        # Update sweep parameters
        self.input_objects[channel]["TIME"].setText(str(sweep_parameters["TIME"]))
        self.input_objects[channel]["RTIME"].setText(str(sweep_parameters["RTIME"]))
        self.input_objects[channel]["FSTART"].setText(str(sweep_parameters["FSTART"]))
        self.input_objects[channel]["FSTOP"].setText(str(sweep_parameters["FSTOP"]))
        self.input_objects[channel]["HTIME_START"].setText(
            str(sweep_parameters["HTIME_START"])
        )
        self.input_objects[channel]["HTIME_STOP"].setText(
            str(sweep_parameters["HTIME_STOP"])
        )

        # Update waveform parameters
        self.input_objects[channel]["FREQUENCY"].setValue(
            waveform_parameters["frequency"]
        )
        self.input_objects[channel]["AMPLITUDE"].setValue(
            waveform_parameters["amplitude"]
        )
        self.input_objects[channel]["OFFSET"].setValue(waveform_parameters["offset"])
        # Assuming WAVEFORM_TYPE is a QComboBox or similar, this line remains unchanged

        self.input_objects[channel]["WAVEFORM_TYPE"].setCurrentText(
            waveform_parameters["waveform_type"]
        )



import abc
from datetime import datetime
from typing import Optional

import pyvisa


class Interface(abc.ABC):

    def __init__(self, resource: pyvisa.Resource, address: Optional[str] = None):
        self.inst = resource
        self.address = address
        self.debug = False

    def write(self, command: str) -> None:
        self.inst.write(command)
        if self.debug:
            print(f"[{datetime.now()}]{command}")

    def read(self, command: str) -> str:
        if self.debug:
            print(f"[{datetime.now()}]{command}")
        return self.inst.query(command)


class EthernetInterface(Interface):
    def __init__(self, resource: pyvisa.Resource):
        # set address
        super().__init__(resource, address=resource.resource_name.split("::")[1])


class USBInterface(Interface):
    def __init__(self, resource: pyvisa.Resource):
        # subject to future change, unsure if resource is supposed to be address as well
        super().__init__(resource, address=resource.resource_name)



from typing import Any, Callable, Dict, Optional, Tuple


class FunctionMap:
    def __init__(self, function_map: dict):
        """
        Initializes the FunctionMap class.

        Args:
            map_file (Optional[Path]): Path to the JSON file that contains the function mapping.
                                       Defaults to 'function_map.json' in the CONFIG directory.
        """
        self.function_map = function_map


    def get_function(self, identifier: str) -> Optional[Callable]:
        """
        Retrieves a function by its identifier.

        Args:
            identifier (str): The identifier of the function.

        Returns:
            Optional[Callable]: The function object if found, otherwise None.
        """
        if identifier in self.function_map:
            module_name, func_name = self.function_map[identifier]
            module = __import__(module_name, globals(), locals(), [func_name], 0)
            return getattr(module, func_name)
        return None

    def add_function(self, identifier: str, func: Callable) -> None:
        """
        Adds a new function to the function map.

        Args:
            identifier (str): The identifier for the function.
            func (Callable): The function object to add.
        """
        self.function_map[identifier] = (func.__module__, func.__name__)
        
    @staticmethod
    def serialize_func(func_data: Tuple[str, str]) -> Dict[str, str]:
        """
        Serializes function data into a JSON serializable format.

        Args:
            func_data (Tuple[str, str]): The module and function name tuple.

        Returns:
            Dict[str, str]: A dictionary representation of the function data.
        """
        module_name, func_name = func_data
        return {"module": module_name, "name": func_name}

    @staticmethod
    def deserialize_func(func_data: Tuple[str, str]) -> Callable:
        """
        Deserializes function data from a tuple format.

        Args:
            func_data (Tuple[str, str]): The module and function name tuple.

        Returns:
            Callable: The deserialized function object.

        Raises:
            ValueError: If the function data format is invalid.
        """
        if isinstance(func_data, tuple) and len(func_data) == 2:
            module_name, func_name = func_data
            module = __import__(module_name, globals(), locals(), [func_name], 0)
            return getattr(module, func_name)
        else:
            raise ValueError(f"Invalid function data format {func_data}")

    def parse_and_call(
        self, func: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Parses arguments and keyword arguments, and calls the given function.

        Args:
            func (Callable): The function to call.
            args (Tuple[Any, ...]): The positional arguments to pass to the function.
            kwargs (Dict[str, Any]): The keyword arguments to pass to the function.

        Returns:
            Any: The result of the function call.
        """
        args = args if args is not None else ()
        kwargs = kwargs if kwargs is not None else {}

        # Extract the argument names for the function
        arg_names = func.__code__.co_varnames[: func.__code__.co_argcount]
        args_dict = dict(zip(arg_names, args))

        # Check for overlapping keys and use values from kwargs if they exist
        for key in args_dict:
            if key not in kwargs:
                kwargs[key] = args_dict[key]

        return func(**kwargs)



from PyQt6.QtGui import QAction, QKeySequence
from PyQt6.QtWidgets import QFileDialog, QMenu, QMenuBar, QMessageBox

from sonaris.frontend.widgets.windows import DeviceWindow, VersionWindow


class MainMenuBar(QMenuBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.initUI()

    def initUI(self):
        # File Menu
        fileMenu = QMenu("&File", self)
        self.version_window = VersionWindow(self.parent())
        self.device_window = DeviceWindow()
        self.addMenu(fileMenu)

        # Save State Action
        saveAct = QAction("&Save State", self)
        saveAct.setShortcut(QKeySequence("Ctrl+S"))
        saveAct.setStatusTip("Saves the program state")
        saveAct.triggered.connect(self.save_state)  # Assuming save_state method exists
        fileMenu.addAction(saveAct)

        # Open State Action
        openAct = QAction("&Open State...", self)
        openAct.setShortcut(QKeySequence("Ctrl+O"))
        openAct.setStatusTip("Opens a saved program state")
        openAct.triggered.connect(self.open_state)
        fileMenu.addAction(openAct)

        # Exit Action
        exitAct = QAction("E&xit", self)
        exitAct.setShortcut(QKeySequence("Alt+F4"))
        exitAct.setStatusTip("Exits the application")
        exitAct.triggered.connect(self.exit_app)  # Assuming exit_app method exists
        fileMenu.addAction(exitAct)
        # Device Menu
        deviceMenu = QMenu("&Device", self)
        self.addMenu(deviceMenu)

        # Device Window Action
        deviceAct = QAction("&Device Window", self)
        deviceAct.setShortcut(QKeySequence("Ctrl+D"))
        deviceAct.setStatusTip("Displays device window")
        deviceAct.triggered.connect(
            self.show_device_window
        )  # Assuming show_device_window method exists
        deviceMenu.addAction(deviceAct)

        # Help Menu
        helpMenu = QMenu("&Help", self)
        self.addMenu(helpMenu)

        # Version Action
        versionAct = QAction("&Version", self)
        versionAct.setStatusTip("Displays version window and developer notes")
        versionAct.triggered.connect(
            self.show_version
        )  # Assuming show_version method exists
        helpMenu.addAction(versionAct)

    def save_state(self):
        # Opens a dialog to save a file
        fileName, _ = QFileDialog.getSaveFileName(
            self, "Save State", "", "State Files (*.state)"
        )
        if fileName:
            self.perform_save(
                fileName
            )  # Assuming perform_save method exists to actually save the state

    def open_state(self):
        # Opens a dialog to open a file
        fileName, _ = QFileDialog.getOpenFileName(
            self, "Open State", "", "State Files (*.state)"
        )
        if fileName:
            self.perform_open(fileName)

    def perform_save(self, fileName):
        # Implement the actual save functionality here
        pass

    def perform_open(self, fileName):
        # Implement the actual open functionality here
        pass

    def exit_app(self):
        # Exit the application
        reply = QMessageBox.warning(
            self,
            "Exit",
            "Application will be closed.",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel,
        )

        if reply == QMessageBox.StandardButton.Yes:
            self.parent().close()  # Or QApplication.instance().quit()

    def show_version(self):
        # Opens the version info window
        self.version_window.show()

    def show_device_window(self):
        self.device_window.show()



import os
import platform
import signal

if platform.system() == "Windows":
    import ctypes

    myappid = "sonaris.sonaris.dummy.string"  # arbitrary string
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
import click
from dotenv import load_dotenv

# Assuming create_app is defined elsewhere in your application
from sonaris.app import create_app, signal_handler
from sonaris.utils.log import get_logger

# Configure logger
logger = get_logger()


def ensure_env_variables():
    """
    Ensure key environment variables are set, using default values if necessary.
    """
    # Load the environment variables from .env file
    load_dotenv()

    # Define the key environment variables to check and their default values
    key_env_vars = ["WORKINGDIR", "DATA", "PYTHONPATH", "ASSETS"]

    # Set environment variables to "" if they are not already set
    for var in key_env_vars:
        if not os.getenv(var):
            os.environ[var] = ""
            continue
        print(f"{var}: {os.getenv(var)}")

@click.group()
def cli():
    """
    Sonaris Command Line Interface.
    """
    pass


def run_application(hardware_mock,grafana):
    """Function to initialize and run the Sonaris application."""
    args_dict = {"hardware_mock": hardware_mock,
                 "grafana": grafana}
    logger.info(args_dict)
    app, window = create_app(args_dict)
    window.show()
    app.exec()



@cli.command()
@click.option("--hardware-mock", "-hm", is_flag=True, help="Run the app in hardware mock mode.")
@click.option("--grafana", is_flag=True, help="Start Grafana container alongside the application. Requires Docker.")
def run(hardware_mock, grafana):
    """Run the Sonaris application."""
    signal.signal(signal.SIGINT, signal_handler)
    try:
        ensure_env_variables()
        logger.info("Running application...")
        run_application(hardware_mock, grafana)
    except KeyboardInterrupt:
        logger.info("Exit signal detected.")

if __name__ == "__main__":
    cli()


from typing import Optional

import numpy as np
from PyQt6.QtCharts import QChart, QLineSeries


def plot_waveform(
    waveform_type: Optional[str] = None,
    frequency: Optional[float] = None,
    amplitude: Optional[float] = None,
    offset: Optional[float] = None,
    params: dict = None,
) -> QChart:
    """
    Generate and plot different types of waveforms.

    Parameters:
        waveform_type (str): Type of waveform to generate ('SIN', 'SQUARE', 'RAMP', 'PULSE', 'NOISE', 'ARB', or 'DC').
        frequency (float): Frequency of the waveform.
        amplitude (float): Amplitude of the waveform.
        offset (float): Offset of the waveform.
        params (dict, optional): Optional dictionary containing waveform parameters. If provided, the individual parameters
                                 will be extracted from this dictionary. Defaults to None.

    Returns:
        figure (plotly.graph_objs.Figure): Plotly figure object containing the generated waveform plot.
    """
    if params is not None:
        # If parameters are provided as a dictionary, extract the values
        frequency = float(params["frequency"])
        amplitude = float(params["amplitude"])
        offset = float(params["offset"])
        waveform_type = str(params["waveform_type"])

    # Generate x values from 0 to 1 with 1000 points
    # Assuming one full period of the waveform, change 1/frequency to display multiple periods if required.
    if isinstance(frequency, float) and frequency == 0.0:
        x_values = np.linspace(0, 1, 1000)
    else:
        x_values = np.linspace(0, 4 / frequency, 1000)
    # Generate y values based on the waveform type
    if waveform_type == "SIN":
        y_values = amplitude * np.sin(2 * np.pi * frequency * x_values) + offset
    elif waveform_type == "SQUARE":
        y_values = (
            amplitude * np.sign(np.sin(2 * np.pi * frequency * x_values)) + offset
        )
    elif waveform_type == "RAMP":
        y_values = (
            amplitude
            * (2 * (x_values * frequency - np.floor(x_values * frequency + 0.5)))
            + offset
        )
    elif waveform_type == "PULSE":
        y_values = amplitude * ((x_values * frequency) % 1 < 0.5) + offset
    elif waveform_type == "NOISE":
        y_values = np.random.normal(0, amplitude, len(x_values)) + offset
    elif waveform_type == "ARB":
        # For an arbitrary waveform, you need to define your own function
        y_values = (
            amplitude * np.sin(2 * np.pi * frequency * x_values) + offset
        )  # Placeholder function
    elif waveform_type == "DC":
        y_values = np.full_like(x_values, amplitude + offset)
    else:
        y_values = np.zeros_like(x_values)

    return x_values, y_values


def plot_sweep(
    start_frequency: Optional[float] = None,
    stop_frequency: Optional[float] = None,
    duration: Optional[float] = None,
    rtime: Optional[float] = None,
    htime_start: Optional[float] = None,
    htime_stop: Optional[float] = None,
    params: dict = None,
) -> QChart:
    """
    Generate and plot a frequency sweep.

    Parameters:
        start_frequency (float): Starting frequency of the sweep.
        stop_frequency (float): Stopping frequency of the sweep.
        duration (float): Duration of the sweep.
        rtime (float, optional): Return Time for the sweep.
        htime_start (float, optional): Hold Time Start for the sweep.
        htime_stop (float, optional): Hold Time End for the sweep.
        params (dict, optional): Optional dictionary containing sweep parameters. If provided, the individual parameters
                                 will be extracted from this dictionary. Defaults to None.

    Returns:
        figure (plotly.graph_objs.Figure): Plotly figure object containing the generated sweep plot.
    """
    if params is not None:
        # If parameters are provided as a dictionary, extract the values
        start_frequency = float(params.get("FSTART", start_frequency))
        stop_frequency = float(params.get("FSTOP", stop_frequency))
        duration = float(params.get("TIME", duration))
        rtime = float(params.get("RTIME", rtime))
        htime_start = float(params.get("HTIME_START", htime_start))
        htime_stop = float(params.get("HTIME_STOP", htime_stop))

    # Calculate the number of samples based on the provided duration
    num_samples = int(1000 * duration)  # Convert to samples

    # Ensure that the number of samples is non-negative
    if num_samples < 0:
        num_samples = 0

    # Generate time values based on the number of samples
    t_values = np.linspace(0, duration, num_samples)

    # Apply Hold Time
    if htime_start is not None and htime_stop is not None:
        num_samples_start = int(1000 * htime_start)
        num_samples_hold = int(1000 * (htime_stop - htime_start))
        num_samples_end = int(1000 * (duration - htime_stop))

        # Ensure non-negative number of samples
        num_samples_start = max(num_samples_start, 0)
        num_samples_hold = max(num_samples_hold, 0)
        num_samples_end = max(num_samples_end, 0)

        # Generate time values for hold time
        t_values = np.concatenate(
            [
                np.linspace(0, htime_start, num_samples_start),
                np.linspace(htime_start, htime_stop, num_samples_hold),
                np.linspace(htime_stop, duration, num_samples_end),
            ]
        )

    # Generate a linearly increasing frequency array
    frequency_values = np.linspace(start_frequency, stop_frequency, len(t_values))

    # Generate y values based on the time-varying frequency
    y_values = np.sin(
        2 * np.pi * np.cumsum(frequency_values) * np.mean(np.diff(t_values))
    )

    # Create a plotly figure with t (time) and y values
    series = QLineSeries()
    for t, y in zip(t_values, y_values):
        series.append(t, y)
    return t_values, y_values



import logging
import os
import time
from datetime import timedelta
from pathlib import Path

from filelock import FileLock

from sonaris.defaults import STATE_FILE
from sonaris.utils import log as logutils

# Setting up basic logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class StateManager:
    def __init__(self, json_file: Path = None):
        self.json_file = json_file or STATE_FILE
        self.lock_file = self.json_file.with_suffix(".lock")
        self.data = self.default_state()
        self.birthdate = time.time()

    def sanitize_key(self, key: str) -> str:
        """
        Sanitize the key by stripping leading and trailing spaces and replacing spaces with underscores.
        """
        return key.strip().replace(" ", "_")

    def read_state(self) -> dict:
        with FileLock(self.lock_file, timeout=10):
            # Utilize the load_json_with_backup utility function with locking
            self.data = (
                logutils.load_json_with_backup(self.json_file) or self.default_state()
            )
        return self.data

    def write_state(self, state: dict):
        with FileLock(self.lock_file, timeout=10):
            self.data.update(state)
            logutils.save_json(self.data, self.json_file)

    def default_state(self):
        return {}

    def update_device_last_alive(self, device_type: str, last_alive_time=None):
        state = self.read_state()
        sanitized_key = self.sanitize_key(f"{device_type}_last_alive")
        state[sanitized_key] = last_alive_time or time.time()
        self.write_state(state)

    def get_device_last_alive(self, device_type: str):
        state = self.read_state()
        sanitized_key = self.sanitize_key(f"{device_type}_last_alive")
        return state.get(sanitized_key)

    def get_uptime(self) -> timedelta:
        uptime_seconds = time.time() - self.birthdate
        return timedelta(seconds=int(uptime_seconds))



from datetime import datetime, timedelta
from enum import Enum

from sonaris import factory
from sonaris.defaults import DeviceName
from sonaris.device.dg4202 import DG4202
from sonaris.tasks.task_decorator import parameter_annotations, parameter_constraints

"""
These tasks are used by the scheduler, they are wrappers for the scheduler to call the manager objects.
You will have to point to them under header.py
"""


class TaskName(Enum):
    DG4202_TOGGLE = "Toggle Output"
    DG4202_SET_WAVEFORM = "Set Waveform Parameters"
    DG4202_SET_SWEEP = "Set Sweep Parameters"
    EDUX1002A_AUTO = "Press Auto"


@parameter_constraints(channel=(1, 2), output=["ON", "OFF"])
def task_on_off_dg4202(channel: int, output: bool) -> bool:
    factory.dg4202_manager.device.output_on_off(
        channel=channel,
        status=output,  # the decorator above will be handled by ui_factory.py
    )
    return True


@parameter_annotations(
    frequency="Hz",
    offset="V",
    amplitude="V",
)
@parameter_constraints(
    frequency=(0.0, DG4202.FREQ_LIMIT),
    channel=[1, 2],  # decorating with list means forced options
    waveform_type=DG4202.available_waveforms(),
    offset=(0.0, 5.0),
    amplitude=(0.0, 5.0),
)
def task_set_waveform_parameters(
    channel: int,
    send_on: bool,
    waveform_type: str,
    amplitude: float,
    frequency: float,
    offset: float,
) -> bool:
    factory.dg4202_manager.device.set_waveform(
        channel=channel,
        waveform_type=waveform_type,
        amplitude=amplitude,
        frequency=frequency,
        params=None,
        offset=offset,
    )
    if send_on:
        factory.dg4202_manager.device.output_on_off(channel, True)
    return True


@parameter_annotations(
    fstart="Hz",
    fstop="Hz",
    time="s",
    rtime="ms",
    htime_start="ms",
    htime_stop="ms",
)
@parameter_constraints(
    channel=[1, 2],
    fstart=(0.0, DG4202.FREQ_LIMIT),
    fstop=(0.0, DG4202.FREQ_LIMIT),
    time=(0.0, float("inf")),
    rtime=(0.0, float("inf")),
    htime_start=(0.0, float("inf")),
    htime_stop=(0.0, float("inf")),
)
def task_set_sweep_parameters(
    channel: int,
    send_on: bool,
    fstart: float,
    fstop: float,
    time: float,
    rtime: float = 0,
    htime_start: float = 0,
    htime_stop: float = 0,
) -> bool:
    params = {
        "FSTART": fstart,
        "FSTOP": fstop,
        "TIME": time,
        "RTIME": rtime,
        "HTIME_START": htime_start,
        "HTIME_STOP": htime_stop,
    }
    factory.dg4202_manager.device.set_sweep_parameters(
        channel=channel, sweep_params=params
    )
    if send_on:
        factory.dg4202_manager.device.output_on_off(channel, True)
    return True


@parameter_constraints(press=["OK"])
def task_auto_edux1002a(press: str):
    # for testing, kwarg_value means nothing
    factory.edux1002a_manager.device.autoscale()
    return True


"""
The task list is read by the job_scheduler module and app.py to register the task and render the UI.
"""

TASK_LIST_DICTIONARY = {
    DeviceName.DG4202.value: {
        TaskName.DG4202_TOGGLE.value: task_on_off_dg4202,
        TaskName.DG4202_SET_WAVEFORM.value: task_set_waveform_parameters,
        TaskName.DG4202_SET_SWEEP.value: task_set_sweep_parameters,
    },
    DeviceName.EDUX1002A.value: {TaskName.EDUX1002A_AUTO.value: task_auto_edux1002a},
}


def get_tasks(flatten: bool = False) -> dict:
    """Returns the dict of { device : { task-name : func_pointer , ..} ..}

    Returns:
        dict: dictionary containing devices and its tasks.
    """
    if flatten:
        return {
            inner_key: value
            for outer_dict in TASK_LIST_DICTIONARY.values()
            for inner_key, value in outer_dict.items()
        }
    return TASK_LIST_DICTIONARY


def calculate_schedule_times(steps):
    """
    Calculate the schedule times for each step based on wait and at_time keys.
    :param steps: A list of step dictionaries from the experiment configuration.
    :return: A list of tuples (step, schedule_time) with calculated schedule times.
    """
    last_schedule_time = datetime.now()  # Initialize with the current time
    schedule_times = []

    for step in steps:
        if "at_time" in step:
            at_time = timedelta(seconds=step["at_time"])
            schedule_time = datetime.now() + at_time
        else:
            wait_time = timedelta(seconds=step.get("wait", 0))
            schedule_time = last_schedule_time + wait_time

        schedule_times.append((step, schedule_time))
        last_schedule_time = schedule_time  # Update for the next iteration

    return schedule_times



import argparse
import os
import signal
from pathlib import Path
from typing import Dict, Tuple

os.environ["PYQTGRAPH_QT_LIB"] = "PyQt6"
from importlib.resources import files
from importlib.resources import path as resource_path

import pyvisa
import qdarktheme
from PyQt6.QtCore import QLocale
from PyQt6.QtGui import QGuiApplication, QIcon
from PyQt6.QtWidgets import QApplication, QStackedWidget, QWidget

from sonaris import factory
from sonaris.defaults import (
    APP_NAME,
    DEFAULT_DATADIR,
    GF_DASHBOARDS_DIR,
    GF_DATASOURCES_DIR,
    MONITOR_FILE,
    OSCILLOSCOPE_BUFFER_SIZE,
    TIMEKEEPER_JOBS_FILE,
    VERSION_STRING,
    DeviceName,
)
from sonaris.frontend.managers.dg4202 import DG4202Manager
from sonaris.frontend.managers.edux1002a import EDUX1002AManager
from sonaris.frontend.managers.state_manager import StateManager
from sonaris.frontend.pages.general import GeneralPage
from sonaris.frontend.pages.monitor import MonitorPage
from sonaris.frontend.pages.scheduler import SchedulerPage
from sonaris.frontend.pages.settings import SettingsPage
from sonaris.frontend.widgets.menu import MainMenuBar
from sonaris.frontend.widgets.sidebar import Sidebar
from sonaris.frontend.widgets.templates import ModularMainWindow
from sonaris.scheduler import registry
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.scheduler.worker import Worker
from sonaris.services.datasource import DataSourceService
from sonaris.services.grafana import GrafanaService
from sonaris.tasks.tasks import get_tasks
from sonaris.utils.container import client, network
from sonaris.utils.log import get_logger

logger = get_logger()

# Before creating your application instance
QLocale.setDefault(QLocale(QLocale.Language.English, QLocale.Country.UnitedStates))

logger.info(f"{APP_NAME} {VERSION_STRING}")
logger.info(f"Using {DEFAULT_DATADIR} as working directory.")
logger.info(f"Using {TIMEKEEPER_JOBS_FILE} as persistence file.")
logger.info(f"Using {OSCILLOSCOPE_BUFFER_SIZE} oscilloscope buffer size.")
logger.info(f"Device events under {MONITOR_FILE}.")


def signal_handler(signum, frame):
    logger.info("Exit signal detected.")
    QApplication.quit()
    factory.grafana_service.stop()
    factory.datasource_service.stop()
    factory.worker.stop_worker()
    # Invoke the default SIGINT handler to exit the application
    signal.signal(signum, signal.SIG_DFL)
    os.kill(os.getpid(), signum)


def init_objects(args_dict: dict):
    # ================= Hardware Managers===================#
    factory.resource_manager = pyvisa.ResourceManager()
    factory.state_manager = StateManager()
    factory.edux1002a_manager = EDUX1002AManager(
        state_manager=factory.state_manager,
        args_dict=args_dict,
        resource_manager=factory.resource_manager,
        buffer_size=OSCILLOSCOPE_BUFFER_SIZE,
    )
    factory.dg4202_manager = DG4202Manager(
        factory.state_manager,
        args_dict=args_dict,
        resource_manager=factory.resource_manager,
    )
    factory.worker = Worker(
        function_map=registry.function_map,
        logger=logger,
    )
    factory.timekeeper = Timekeeper(
        persistence_file=TIMEKEEPER_JOBS_FILE,
        worker_instance=factory.worker,
        logger=logger,
    )

    # ================= Register Tasks ===================#
    for task_name, func_pointer in get_tasks(flatten=True).items():
        factory.worker.register_task(func_pointer, task_name)
    # ==================== Services ======================#
    if args_dict["grafana"]:
        factory.grafana_service = GrafanaService(
            client=client, port=None  # Use default client instance  # Use default port
        )
        factory.datasource_service = DataSourceService(
            timekeeper=factory.timekeeper,
            port=None,  # Use default port
            logger=logger,
            name=f"{APP_NAME}DataSource",  # Customize as needed
        )
        factory.datasource_service.write_provisioning_files(
            dashboards_dir=GF_DASHBOARDS_DIR, datasources_dir=GF_DATASOURCES_DIR
        )
        factory.datasource_service.start()
        factory.grafana_service.start()
    factory.worker.start_worker()


class MainWindow(ModularMainWindow):
    def __init__(self, args_dict: dict) -> None:
        super().__init__()
        menu_bar = MainMenuBar(self)
        self.setWindowTitle(APP_NAME)
        self.setMenuBar(menu_bar)
        self.last_page = ""
        # ---------------------------------------------------------------------- #
        # ---------------------------SIDEBAR SETUP------------------------------ #
        self.sidebar = Sidebar(self)
        device_managers = {
            DeviceName.DG4202.value: factory.dg4202_manager,
            DeviceName.EDUX1002A.value: factory.edux1002a_manager,
        }
        self.sidebar.sizePolicy()
        self.sidebar_dict: Dict[str, QWidget] = {
            GeneralPage.PAGE_NAME: GeneralPage(
                dg4202_manager=factory.dg4202_manager,
                edux1002a_manager=factory.edux1002a_manager,
                parent=self,
                args_dict=args_dict,
                root_callback=self.root_callback,
            ),
            SchedulerPage.PAGE_NAME: SchedulerPage(
                timekeeper=factory.timekeeper,
                parent=self,
                args_dict=args_dict,
                root_callback=self.root_callback,
            ),
            MonitorPage.PAGE_NAME: MonitorPage(
                device_managers=device_managers,
                parent=self,
                args_dict=args_dict,
                monitor_logs=MONITOR_FILE,
                root_callback=self.root_callback,
            ),
            SettingsPage.PAGE_NAME: SettingsPage(
                device_managers=device_managers,
                parent=self,
                args_dict=args_dict,
                root_callback=self.root_callback,
            ),
        }
        self.sidebar.addItems(self.sidebar_dict.keys())  # Add strings to sidebar items
        self.sidebar_content = QStackedWidget(self)
        list(
            map(self.sidebar_content.addWidget, self.sidebar_dict.values())
        )  # Adds all child widgets to content widgets

        # --------------------------------------------------------------------- #
        # ---------------------------LAYOUT SETUP------------------------------ #
        # --------------------------------------------------------------------- #

        self.add_widget_to_left(self.sidebar)
        self.add_widget_to_middle(
            self.sidebar_content
        )  # Use the method from ModularMainWindow
        # Connect the Sidebar's custom signal to the MainWindow's slot
        self.sidebar.pageSelected.connect(self.loadPage)

    def loadPage(self, page_name: str) -> None:
        page_widget: QWidget = self.sidebar_dict.get(page_name)
        if page_widget and self.last_page != page_name:
            self.sidebar_content.setCurrentWidget(page_widget)
            # logger.info(f"[Sidebar] Switch event to {page_name}")
            page_widget.update()
            self.last_page = page_name

    def root_callback(self):
        """tells pages the pages to call update method."""
        for _, page_obj in self.sidebar_dict.items():
            page_obj.update()

    def closeEvent(self, event):
        logger.info("main x exit button was clicked")
        factory.worker.stop_worker()
        super().closeEvent(event)

    def showEvent(self, event):
        primaryScreen = QGuiApplication.primaryScreen()
        screenGeometry = primaryScreen.availableGeometry()
        centerPoint = screenGeometry.center()
        frameGeometry = self.frameGeometry()
        frameGeometry.moveCenter(centerPoint)
        self.move(frameGeometry.topLeft())
        super().showEvent(event)


def create_app(args_dict: dict) -> Tuple[QApplication, MainWindow]:
    init_objects(args_dict=args_dict)

    app = QApplication([])
    window = MainWindow(args_dict)

    qdarktheme.setup_theme()

    # Assuming 'sonaris.assets' is a valid package path
    favicon_path = files("sonaris.assets").joinpath("favicon.ico")
    app_icon = QIcon(str(favicon_path))

    app.setWindowIcon(app_icon)
    window.setWindowIcon(app_icon)

    window.setWindowTitle(APP_NAME)
    window.resize(640, 400)
    logger.info(f"Window size after resize: {window.size()} ")
    return app, window


def run_application(args_dict):
    logger.info(args_dict)
    app, window = create_app(args_dict)
    window.show()
    app.exec()


if __name__ == "__main__":
    # Not recommended anymore to run this way. (sonaris/__main__.py controls the argument parsing using click)
    # Do python -m sonaris run --hardware-mock to run the app in hardware mock mode instead ( or poetry run python -m sonaris run --hardware-mock)
    # You can do packaged runs as well by doing 'pip install .'
    parser = argparse.ArgumentParser(description="Run the sonaris application.")
    parser.add_argument(
        "--hardware-mock",
        action="store_true",
        help="Run the app in hardware mock mode.",
    )
    args = parser.parse_args()
    args_dict = vars(args)
    run_application(args_dict)



import os
from enum import Enum
from pathlib import Path

APP_NAME = "Sonaris"
VERSION_STRING = "v0.1.4"

GRAPH_RGB = (255, 255, 255)
OSCILLOSCOPE_BUFFER_SIZE = 512


class ErrorLevel(Enum):
    INFO = 0  # Missing parameters, etc.
    BAD_CONFIG = 1  # Missing tasks, etc.
    INVALID_YAML = 2  # Issues with loading/parsing YAML


class DeviceName(Enum):
    DG4202 = "DG4202"
    EDUX1002A = "EDUX1002A"


DEVICE_LIST = [DeviceName.DG4202.value, DeviceName.EDUX1002A.value]

DEFAULT_DATADIR = Path().home() / f".{APP_NAME.lower()}"
DEFAULT_DATADIR.mkdir(parents=True, exist_ok=True)

# Assuming DEFAULT_DATADIR is already a Path object
DEFAULT_DATADIR = Path(os.getenv("DATA")) if os.getenv("DATA") else DEFAULT_DATADIR
# Define file paths in one-liners, checking for existence and falling back to DEFAULT_DATADIR if necessary
STATE_FILE = (
    Path(DEFAULT_DATADIR / "state.json").exists()
    and Path(DEFAULT_DATADIR / "state.json")
    or (DEFAULT_DATADIR / "state.json")
)
TIMEKEEPER_JOBS_FILE = (
    Path(DEFAULT_DATADIR / "jobs.json").exists()
    and Path(DEFAULT_DATADIR / "jobs.json")
    or (DEFAULT_DATADIR / "jobs.json")
)
MONITOR_FILE = (
    Path(DEFAULT_DATADIR / "monitor.json").exists()
    and Path(DEFAULT_DATADIR / "monitor.json")
    or (DEFAULT_DATADIR / "monitor.json")
)
SETTINGS_FILE = (
    Path(DEFAULT_DATADIR / "settings.json").exists()
    and Path(DEFAULT_DATADIR / "settings.json")
    or (DEFAULT_DATADIR / "settings.json")
)
LOG_DIR = (
    Path(DEFAULT_DATADIR / "logs").exists()
    and Path(DEFAULT_DATADIR / "logs")
    or (DEFAULT_DATADIR / "logs")
)
LOG_DIR.mkdir(parents=True, exist_ok=True)
# UI CONFIG
TICK_INTERVAL = 500.0  # in ms
DECIMAL_POINTS = 5
NOT_FOUND_STRING = "Device not found!"
WAIT_KEYWORD = "wait"
TIMESTAMP_KEYWORD = "timestamp"
TASKS_MISSING = "No tasks available"
AT_TIME_KEYWORD = "at_time"
EXPERIMENT_KEYWORD = "experiment"
DELAY_KEYWORD = "delay"
# GRAFANA DEFAULTS
DEFAULT_TAB_STYLE = {"height": "30px", "padding": "2px"}
GF_SECURITY_ADMIN_PASSWORD = os.getenv("GF_SECURITY_ADMIN_PASSWORD", "admin")
GF_PORT = int(os.getenv("GF_PORT", "3000"))
# sonaris.defaults.py
GF_INSTALL_PLUGINS = "grafana-simple-json-datasource,yesoreyeram-infinity-datasource"


DATA_SOURCE_PORT = int(os.getenv("DATA_SOURCE_PORT", "5000"))
DATA_SOURCE_NAME = str(os.getenv("DATA_SOURCE_NAME", f"{APP_NAME}DataSource"))

GF_DATASOURCES_DIR = DEFAULT_DATADIR / "provisioning" / "datasources"
GF_DASHBOARDS_DIR = DEFAULT_DATADIR / "provisioning" / "dashboards"

SONARIS_NETWORK_NAME = f"{APP_NAME.lower()}_network"
# GRAFANA DEFAULTS



from typing import Any, Callable, Tuple, Type

from PyQt6.QtCore import pyqtBoundSignal
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QLineEdit,
    QSpinBox,
    QWidget,
)

from sonaris.defaults import DECIMAL_POINTS


class UIComponentFactory:

    TRUTH_VALUES = [
        "true",
        "1",
        "on",
        "ok",
        "yes",
        "resume",
        "continue",
    ]
    FALSE_VALUES = [
        "false",
        "0",
        "off",
        "no",
        "cancel",
        "abort",
        "deny",
    ]

    SIGNAL_MAP = {
        QLineEdit: "textChanged",
        QSpinBox: "valueChanged",
        QDoubleSpinBox: "valueChanged",
        QComboBox: "currentIndexChanged",
        QCheckBox: "stateChanged",
    }

    @staticmethod
    def create_widget(
        parameter_name: str,
        value: Any,
        expected_type: type,
        constraints: Any,
        callback: Callable = None,
    ) -> QWidget:
        """
        Creates a widget based on the parameter's expected type and the specific constraints.
        """
        widget, default_value = UIComponentFactory.map_type_to_widget(
            expected_type, constraints
        )

        # Apply the current value or default to the widget
        if isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            widget.setValue(
                expected_type(value)
                if value is not None
                else expected_type(default_value)
            )
        elif isinstance(widget, QCheckBox):
            widget.setChecked(
                value if value is not None else expected_type(default_value)
            )
        elif isinstance(widget, QComboBox):
            if expected_type == bool:
                # Handle boolean parameters differently
                if isinstance(value, str):
                    index = UIComponentFactory.findTextIndexForBooleanValue(
                        widget, value
                    )
                    widget.setCurrentIndex(index)
                elif isinstance(value, bool):
                    # Convert boolean to string representation to find the index
                    value_str = "true" if value else "false"
                    index = UIComponentFactory.findTextIndexForBooleanValue(
                        widget, value_str
                    )
                    widget.setCurrentIndex(index)
            elif value in constraints:  # Assuming `constraints` is a list for QComboBox
                index = widget.findData(value)
                widget.setCurrentIndex(index)
            else:
                widget.setCurrentIndex(0)
        elif isinstance(widget, QLineEdit):
            widget.setText(str(value) if value is not None else str(default_value))
        widget.setProperty("expected_type", expected_type.__name__)
        widget.setProperty("parameter_name", parameter_name)
        if callback:
            UIComponentFactory.connect_widget_signal(widget, callback)
        return widget

    @staticmethod
    def findTextIndexForBooleanValue(widget: QComboBox, value: str) -> int:
        # Normalize the string value for comparison
        normalized_value = value.lower()
        # Determine if the value is considered true or false
        if normalized_value in UIComponentFactory.TRUTH_VALUES:
            truthy = True
        elif normalized_value in UIComponentFactory.FALSE_VALUES:
            truthy = False
        else:
            return 0  # Default to the first index if undetermined

        # Iterate through the combo box items to find the match
        for i in range(widget.count()):
            item_text = widget.itemText(i).lower()
            if (truthy and item_text in UIComponentFactory.TRUTH_VALUES) or (
                not truthy and item_text in UIComponentFactory.FALSE_VALUES
            ):
                return i
        return 0  # Default to the first index if no match found

    @staticmethod
    def map_type_name_to_type(type_name: str):
        """
        Maps a type name (string) back to a type. This is needed for casting values fetched from UI components.
        """
        return {
            "bool": bool,
            "int": int,
            "float": float,
            "str": str,
        }.get(
            type_name, None
        )  # Default to str if not found

    @staticmethod
    def map_type_to_widget(
        param_type: type, constraints: Any = None
    ) -> Tuple[QWidget, Any]:
        """
        Maps a parameter type to a PyQt widget, considering the constraints.
        """
        if param_type in [int, float, str, bool] and isinstance(constraints, list):
            widget = QComboBox()
            for value in constraints:
                widget.addItem(str(value), value)
            default_value = constraints[0]
        elif param_type == int:
            widget = QSpinBox()
            if constraints and isinstance(constraints, tuple):
                widget.setMinimum(constraints[0])
                widget.setMaximum(constraints[1])
            else:
                widget.setRange(-2147483648, 2147483647)  # Default 32-bit integer range
            default_value = 0

        elif param_type == float:
            widget = QDoubleSpinBox()
            widget.setDecimals(DECIMAL_POINTS)
            if constraints and isinstance(constraints, tuple):
                widget.setMinimum(constraints[0])
                widget.setMaximum(constraints[1])
            else:
                widget.setRange(-1.0e100, 1.0e100)
                # Default to 3 decimal places for broad applicability
            default_value = 0.0

        elif param_type == str:
            widget = QLineEdit()
            # Optionally set a placeholder text here to guide the user
            default_value = ""

        elif param_type == bool:
            widget = QCheckBox()
            default_value = False
        else:  # Default case, especially for types like str without specific constraints
            widget = QLineEdit()
            default_value = ""

        # Set property to identify the expected type easily
        return widget, default_value

    @staticmethod
    def connect_widget_signal(widget: QWidget, callback: Callable[[Any], None]) -> None:
        """
        Connects the appropriate signal of the widget to the given callback function.

        Args:
            widget: The widget whose signal needs to be connected.
            callback: The callback function to be called when the signal is emitted.
        """
        widget_class: Type[QWidget] = type(widget)  # Get the class of the widget
        if widget_class in UIComponentFactory.SIGNAL_MAP:
            signal_name: str = UIComponentFactory.SIGNAL_MAP[widget_class]
            signal: pyqtBoundSignal = getattr(widget, signal_name)
            signal.connect(callback)
        else:
            raise ValueError("Unsupported widget type for signal connection")

    @staticmethod
    def extract_value(widget: QWidget, cast_type: type = None) -> Any:
        """Extracts and returns the value from the widget, casting it to the expected type."""
        if cast_type is None:
            cast_type = UIComponentFactory.map_type_name_to_type(
                widget.property("expected_type")
            )

        # From here cast_type is determined
        if isinstance(widget, QLineEdit):
            value = widget.text()
        elif isinstance(widget, QCheckBox):
            value = widget.isChecked()
        elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            value = widget.value()
        elif isinstance(widget, QComboBox):
            value = widget.currentData()
        else:
            value = None

        # Special handling for boolean cast_type
        if cast_type == bool:
            if isinstance(value, str):
                # Normalize the string to lower case to simplify comparisons
                return value.lower() in UIComponentFactory.TRUTH_VALUES
            else:
                return bool(
                    value
                )  # Directly cast to bool for non-string values depending of Pythonic truthiness
        elif cast_type in [int, float, str]:
            return cast_type(value) if cast_type else None
        return value if value else None



from abc import ABC, abstractmethod
from typing import NoReturn
from uvicorn.server import Server,Config
import contextlib
import time
from fastapi import FastAPI
import threading
class Service(ABC):
    def __init__(self):
        pass
    @abstractmethod
    def start(self) -> NoReturn:
        """
        Start the service, creating or finding necessary resources.
        """
        raise NotImplementedError
    
    @abstractmethod
    def stop(self) -> NoReturn:
        """
        Stop the service and clean up resources if necessary.
        """
        raise NotImplementedError

class MultithreadedServer:
    def __init__(self, app: FastAPI, host: str = "0.0.0.0", port: int = 8000, log_level: str = "info"):
        self.config = Config(app=app, host=host, port=port, log_level=log_level)
        self.server = Server(config=self.config)
        self.thread = threading.Thread(target=self.server.run, args=(), daemon=True)

    def start(self):
        self.thread.start()

    def stop(self):
        # Uvicorn doesn't provide a direct way to stop the server from another thread, but setting `should_exit` helps in stopping the loop.
        self.server.should_exit = True
        self.thread.join()


from typing import Callable

from PyQt6.QtWidgets import QLabel, QVBoxLayout

from sonaris.defaults import TICK_INTERVAL
from sonaris.frontend.managers.dg4202 import DG4202Manager
from sonaris.frontend.managers.edux1002a import EDUX1002AManager
from sonaris.frontend.widgets.gen_oscilloscope import EDUX1002AOscilloscopeWidget
from sonaris.frontend.widgets.gen_signal import DG4202SignalGeneratorWidget
from sonaris.frontend.widgets.templates import BasePage


class GeneralPage(BasePage):
    PAGE_NAME = "General"

    def __init__(
        self,
        dg4202_manager: DG4202Manager,
        edux1002a_manager: EDUX1002AManager,
        parent=None,
        args_dict: dict = None,
        root_callback: Callable = None,
    ):
        # Call the constructor of the BasePage class
        super().__init__(
            parent=parent, args_dict=args_dict, root_callback=root_callback
        )

        # Initialize GeneralPage-specific attributes
        self.dg4202_manager = dg4202_manager
        self.edux1002a_manager = edux1002a_manager

        self.initUI()

    def check_connection(self) -> bool:
        return self.dg4202_manager.is_device_alive()

    def initUI(self):
        self.main_layout = QVBoxLayout()
        self.oscilloscope = EDUX1002AOscilloscopeWidget(
            self.edux1002a_manager, tick=TICK_INTERVAL
        )
        self.main_layout.addWidget(self.oscilloscope)
        self.status_label = QLabel("")
        self.main_layout.addWidget(self.status_label)

        self.default_widget = DG4202SignalGeneratorWidget(self.dg4202_manager, self)
        self.main_layout.addWidget(self.default_widget)

        self.setLayout(self.main_layout)

    def update(self):
        if self.check_connection():
            self.default_widget.update()



import json
from datetime import datetime
from pathlib import Path
from typing import Dict

from PyQt6 import QtCore
from PyQt6.QtCore import QTimer
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QHeaderView,
    QListWidget,
    QMessageBox,
    QPushButton,
    QSplitter,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from sonaris.defaults import MONITOR_FILE
from sonaris.frontend.managers.device import DeviceManager
from sonaris.utils import log as logutils

logger = logutils.get_logger()


class DeviceMonitorWidget(QWidget):
    def __init__(
        self,
        device_managers: Dict[str, DeviceManager],
        monitor_logs: Path = None,
        parent=None,
    ):
        super().__init__(parent)
        self.device_managers = device_managers  # Dictionary of device managers
        self.monitor_logs = monitor_logs or MONITOR_FILE
        self.device_statuses = {
            device_name: False for device_name in self.device_managers.keys()
        }
        self.initUI()
        self.setup_refresh_timer()
        self.load_event_log()

    def initUI(self):
        self.splitter = QSplitter(self)
        self.leftWidget = QWidget()
        self.leftLayout = QVBoxLayout(self.leftWidget)

        # Device Status Table
        self.status_table = QTableWidget(0, 2)  # Start with zero rows and two columns
        self.status_table.setHorizontalHeaderLabels(["Device Name", "Status"])
        # Make the table columns stretch to fit the space
        self.status_table.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.status_table.horizontalHeader().setStretchLastSection(True)

        self.leftLayout.addWidget(self.status_table)

        # Right Side Widget and Layout
        self.rightWidget = QWidget()
        self.rightLayout = QVBoxLayout(self.rightWidget)
        self.event_log_list = QListWidget()
        self.rightLayout.addWidget(self.event_log_list)

        self.splitter.addWidget(self.leftWidget)
        self.splitter.addWidget(self.rightWidget)

        # Clear List Button
        self.clear_list_button = QPushButton("Clear Event Log")
        self.clear_list_button.clicked.connect(self.clear_event_log)
        self.rightLayout.addWidget(self.clear_list_button)

        self.mainLayout = QHBoxLayout(self)
        self.mainLayout.addWidget(self.splitter)

    def setup_refresh_timer(self):
        self.refresh_timer = QTimer(self)
        self.refresh_timer.timeout.connect(self.update_device_statuses)
        self.refresh_timer.start(1000)  # Refresh every 1 second

    def load_event_log(self):

        event_log = logutils.load_json_with_backup(self.monitor_logs)

        for event in event_log:
            self.event_log_list.addItem(
                f"[{event['timestamp']}] - {event['description']}"
            )

    def clear_event_log(self):
        # Confirmation message box
        reply = QMessageBox.question(
            self,
            "Clear Archive",
            "Are you sure you want to clear the monitoring logs?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )

        # Check if the user confirmed the action
        if reply == QMessageBox.StandardButton.Yes:
            try:
                self.event_log_list.clear()
                with open(self.monitor_logs, "w") as file:
                    json.dump([], file, indent=4)
                QMessageBox.information(
                    self, "Success", "The monitoring logs has been cleared."
                )
            except Exception as e:
                QMessageBox.critical(self, "Error", f"Could not clear the archive. {e}")

    def log_event(self, description: str):
        timestamp = datetime.now().isoformat()
        logger.info(f"[Monitor]: {description}")
        event = {"timestamp": timestamp, "description": description}
        if not self.monitor_logs.exists():
            with open(self.monitor_logs, "w") as file:
                json.dump([event], file, indent=4)
        else:
            with open(self.monitor_logs, "r+") as file:
                event_log = json.load(file)
                event_log.append(event)
                file.seek(0)
                json.dump(event_log, file, indent=4)
        self.event_log_list.addItem(f"[{timestamp}] - {description}")

    def update_device_statuses(self):
        self.status_table.setRowCount(len(self.device_managers))
        for row, (device_name, manager) in enumerate(self.device_managers.items()):
            self.update_device_status(device_name, manager, row)
 
    def update_device_status(self, device_name: str, manager: DeviceManager, row: int):
        try:
            is_alive = manager.is_device_alive()
            if is_alive and self.device_statuses[device_name] != is_alive:
                self.log_event(f"{device_name} Connected")
            elif not is_alive and self.device_statuses[device_name] != is_alive:
                self.log_event(f"{device_name} Disconnected")

            self.device_statuses[device_name] = is_alive
            status = "Connected" if is_alive else "Disconnected"

            # Create QTableWidgetItem instances for device name and status
            device_item = QTableWidgetItem(device_name)
            status_item = QTableWidgetItem(status)

            # Make the items non-editable
            device_item.setFlags(device_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)
            status_item.setFlags(status_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            # Set the items in the table
            self.status_table.setItem(row, 0, device_item)
            self.status_table.setItem(row, 1, status_item)
        except Exception as e:
            logger.error(f"Error updating device statuses: {e}")



import json
from typing import Callable

from PyQt6 import QtCore
from PyQt6.QtWidgets import (
    QHeaderView,
    QLabel,
    QMessageBox,
    QPushButton,
    QSplitter,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from sonaris import factory
from sonaris.frontend.widgets.sch_exp_popup import ExperimentConfigPopup
from sonaris.frontend.widgets.sch_task_popup import TaskConfigPopup, TaskDetailsDialog
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.utils import log as logutils
from sonaris.utils.log import get_logger

logger = get_logger()


class SchedulerWidget(QWidget):
    def __init__(self, timekeeper: Timekeeper = None, root_callback: Callable = None):
        super().__init__()
        self.timekeeper = timekeeper or factory.timekeeper
        self.job_popup = TaskConfigPopup(self.timekeeper, self.popup_callback)
        self.experiment_popup = ExperimentConfigPopup(
            self.timekeeper, self.popup_callback
        )
        self.timekeeper.set_callback(self.popup_callback)
        self.root_callback = root_callback
        self.initUI()

    def initUI(self):
        # Create a horizontal splitter
        splitter = QSplitter(QtCore.Qt.Orientation.Horizontal, self)

        # Left side widget and layout
        leftWidget = QWidget()
        leftLayout = QVBoxLayout(leftWidget)

        # Label and table for displaying current jobs
        self.jobsLabel = QLabel("Current Jobs:", leftWidget)
        leftLayout.addWidget(self.jobsLabel)

        # Create the table with 4 columns
        self.jobsTable = QTableWidget(0, 4, leftWidget)
        self.jobsTable.setHorizontalHeaderLabels(
            ["Job ID", "Task Name", "Scheduled Time", "Parameters"]
        )
        self.jobsTable.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.update_jobs_table()
        leftLayout.addWidget(self.jobsTable)

        # Button to configure a job
        self.configureJobButton = QPushButton("Schedule Task", leftWidget)
        self.configureJobButton.clicked.connect(self.open_job_config_popup)
        leftLayout.addWidget(self.configureJobButton)

        # Button to configure a job
        self.configureExperimentButton = QPushButton("Run Experiment", leftWidget)
        self.configureExperimentButton.clicked.connect(
            self.open_experiment_config_popup
        )
        leftLayout.addWidget(self.configureExperimentButton)

        self.removeJobButton = QPushButton("Remove Selected Job", leftWidget)
        self.removeJobButton.clicked.connect(self.remove_selected_job)
        leftLayout.addWidget(self.removeJobButton)

        # Right side widget and layout
        rightWidget = QWidget()
        rightLayout = QVBoxLayout(rightWidget)

        # Label and list for displaying finished jobs
        self.finishedJobsLabel = QLabel("Finished Jobs:", rightWidget)
        rightLayout.addWidget(self.finishedJobsLabel)

        self.finishedJobsTable = QTableWidget(rightWidget)
        self.finishedJobsTable.setColumnCount(
            3
        )  # Assuming you have 3 columns: Result, Task, Job ID
        self.finishedJobsTable.setHorizontalHeaderLabels(["Result", "Task", "Job ID"])
        self.finishedJobsTable.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.finishedJobsTable.setSelectionBehavior(
            QTableWidget.SelectionBehavior.SelectRows
        )
        self.finishedJobsTable.setSelectionMode(
            QTableWidget.SelectionMode.SingleSelection
        )

        self.finishedJobsTable.cellDoubleClicked.connect(self.show_archive_entry)

        self.update_finished_jobs_list()
        rightLayout.addWidget(self.finishedJobsTable)

        self.clearFinishedJobsButton = QPushButton("Clear Archive", rightWidget)
        self.clearFinishedJobsButton.clicked.connect(self.clear_finished_jobs)
        rightLayout.addWidget(self.clearFinishedJobsButton)
        # Add the left and right widgets to the splitter
        splitter.addWidget(leftWidget)
        splitter.addWidget(rightWidget)

        # Optionally set initial sizes of the splitter sections
        splitter.setSizes([300, 200])  # Adjust these values as needed

        # Set the main layout to include the splitter
        mainLayout = QVBoxLayout(self)
        mainLayout.addWidget(splitter)

    def update_jobs_table(self):
        self.jobsTable.setRowCount(0)
        jobs = self.timekeeper.get_jobs()
        for job_id, job_info in jobs.items():
            row_position = self.jobsTable.rowCount()
            self.jobsTable.insertRow(row_position)

            # Create QTableWidgetItem for each entry and set it to non-editable
            job_id_item = QTableWidgetItem(job_id)
            job_id_item.setFlags(
                job_id_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable
            )
            self.jobsTable.setItem(row_position, 0, job_id_item)

            task_name_item = QTableWidgetItem(job_info["task"])
            task_name_item.setFlags(
                task_name_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable
            )
            self.jobsTable.setItem(row_position, 1, task_name_item)

            schedule_time_item = QTableWidgetItem(job_info["schedule_time"])
            schedule_time_item.setFlags(
                schedule_time_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable
            )
            self.jobsTable.setItem(row_position, 2, schedule_time_item)

            parameters_item = QTableWidgetItem(str(job_info["kwargs"]))
            parameters_item.setFlags(
                parameters_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable
            )
            self.jobsTable.setItem(row_position, 3, parameters_item)

    def remove_selected_job(self):
        selected_items = self.jobsTable.selectedItems()
        if selected_items:
            selected_row = selected_items[0].row()
            job_id = self.jobsTable.item(selected_row, 0).text()
            try:
                self.timekeeper.cancel_job(job_id)
                self.update_jobs_table()
            except Exception as e:
                logger.info(f"Error removing job {job_id}: {e}")
        else:
            logger.info("No job selected")

    def update_finished_jobs_list(self):
        self.finishedJobsTable.setRowCount(0)  # Clear the existing rows in the table
        finished_jobs = logutils.load_json_with_backup(self.timekeeper.archive)

        for job_id, job_info in finished_jobs.items():
            row_position = self.finishedJobsTable.rowCount()
            self.finishedJobsTable.insertRow(row_position)

            result_item = QTableWidgetItem(
                "OK" if job_info.get("result", False) else "ERR"
            )
            task_item = QTableWidgetItem(job_info.get("task", ""))
            job_id_item = QTableWidgetItem(job_id)

            result_item.setFlags(
                result_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable
            )
            task_item.setFlags(task_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)
            job_id_item.setFlags(
                job_id_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable
            )

            self.finishedJobsTable.setItem(row_position, 0, result_item)
            self.finishedJobsTable.setItem(row_position, 1, task_item)
            self.finishedJobsTable.setItem(row_position, 2, job_id_item)

    def clear_finished_jobs(self):
        # Confirmation message box
        reply = QMessageBox.question(
            self,
            "Clear Archive",
            "Are you sure you want to clear the archive?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )

        # Check if the user confirmed the action
        if reply == QMessageBox.StandardButton.Yes:
            try:
                self.timekeeper.clear_archive()
                self.update_finished_jobs_list()
                QMessageBox.information(
                    self, "Success", "The archive has been cleared."
                )
            except Exception as e:
                QMessageBox.critical(self, "Error", "Could not clear the archive.")

    def open_job_config_popup(self):
        self.job_popup.exec()

    def open_experiment_config_popup(self):
        self.experiment_popup.exec()

    def popup_callback(self):
        self.update_jobs_table()
        self.update_finished_jobs_list()
        if self.root_callback is not None:
            # Relay tick to main app.
            self.root_callback()

    def show_archive_entry(self, row, column):
        job_id_item = self.finishedJobsTable.item(
            row, 2
        )  # Assuming Job ID is in the 3rd column
        if job_id_item:
            job_id = job_id_item.text()
            try:
                with self.timekeeper.archive.open("r") as file:
                    finished_jobs = json.load(file)
                job_details = finished_jobs.get(job_id, {})
                if job_details:
                    dialog = TaskDetailsDialog(job_details, self)
                    dialog.exec()
            except FileNotFoundError:
                QMessageBox.warning(
                    self, "Error", "File not found. Could not retrieve job details."
                )



from typing import Any, Tuple

import yaml


class ConfigLoader:
    @staticmethod
    def load_config(config_path: str) -> Tuple[bool, str, Any]:
        try:
            with open(config_path, "r") as file:
                config = yaml.safe_load(file)
                if not config:
                    return False, "No configuration loaded", None
                return True, "Configuration loaded successfully", config
        except Exception as e:
            return False, f"Failed to load configuration: {str(e)}", None

    @staticmethod
    def save_config(config: dict, config_path: str) -> Tuple[bool, str]:
        try:
            with open(config_path, "w") as file:
                yaml.dump(config, file, sort_keys=False)
            return True, "Configuration saved successfully."
        except Exception as e:
            return False, f"Failed to save configuration: {str(e)}"


import json
from pathlib import Path

from PyQt6.QtWidgets import QLabel, QLineEdit, QPushButton, QVBoxLayout, QWidget

from sonaris.defaults import VERSION_STRING


class SettingsStateWidget(QWidget):
    def __init__(self, settings_file: Path, parent=None):
        super().__init__(parent)
        self.settings_file = settings_file
        self.load_settings()
        self.initUI()

    def initUI(self):
        layout = QVBoxLayout()

        # Add a label and edit for a sample setting
        self.setting_label = QLabel(f"Sonaris {VERSION_STRING}")
        # self.setting_edit = QLineEdit(self)
        # self.setting_edit.setText(self.settings.get("sample_setting", ""))

        # Create buttons
        # self.apply_button = QPushButton("Apply")
        # self.apply_button.clicked.connect(self.apply_settings)

        # self.reset_button = QPushButton("Reset")
        # self.reset_button.clicked.connect(self.reset_settings)

        # Arrange widgets in the layout
        layout.addWidget(self.setting_label)
        # layout.addWidget(self.setting_edit)
        # layout.addWidget(self.apply_button)
        # layout.addWidget(self.reset_button)

        self.setLayout(layout)

    def load_settings(self):
        try:
            with open(self.settings_file, "r") as f:
                self.settings = json.load(f)
        except (FileNotFoundError, json.JSONDecodeError):
            self.settings = {}

    def save_settings(self):
        with open(self.settings_file, "w") as f:
            json.dump(self.settings, f, indent=4)

    def apply_settings(self):
        # Save the current state of the settings
        self.settings["sample_setting"] = self.setting_edit.text()
        self.save_settings()

    def reset_settings(self):
        # Reset the settings to their saved state
        self.load_settings()
        self.setting_edit.setText(self.settings.get("sample_setting", ""))



import json
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import Optional

import yaml
from fastapi import FastAPI

from sonaris.defaults import (
    DATA_SOURCE_NAME,
    DATA_SOURCE_PORT,
    GF_DASHBOARDS_DIR,
    GF_DATASOURCES_DIR,
)
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.services.service import MultithreadedServer, Service
from sonaris.utils.log import get_logger

logger = get_logger()


class DataSourceService(Service):
    def __init__(
        self,
        timekeeper: Timekeeper,
        port: int = None,
        logger: Optional[Logger] = None,
        name: str = None,
    ):
        super().__init__()
        self.name = str(name) or str(DATA_SOURCE_NAME)
        self.timekeeper = timekeeper
        self.port = port or DATA_SOURCE_PORT
        self.logger = logger or get_logger()
        self.app = FastAPI(title="Sonaris Data Source Service")
        self.setup_routes()

        # Prepare the Uvicorn config here
        self.server = MultithreadedServer(
            app=self.app, port=self.port, log_level="info"
        )

    @staticmethod
    def get_data_source_manifest(
        host="http://host.docker.internal", port=DATA_SOURCE_PORT, name=DATA_SOURCE_NAME
    ):
        # Update the data source type to match the plugin being used (e.g., for Grafana Infinity)
        return {
            "apiVersion": 1,
            "datasources": [
                {
                    "id": 1,
                    "uid": "P3E0B65AA66943F6C",
                    "orgId": 1,
                    "name": "SonarisDataSource",
                    "type": "yesoreyeram-infinity-datasource",
                    "url": f"{host}:{port}",
                    "basicAuth": "false",
                    "basicAuthUser": "",
                    "isDefault": "true",
                    "jsonData": {
                        "global_queries": [
                            {
                                "name": "task archive",
                                "id": "archive",
                                "query": {
                                    "refId": "my-query-1",
                                    "type": "json",
                                    "source": "url",
                                    "data": "",
                                    "root_selector": "",
                                    "columns": [],
                                    "filters": [],
                                    "format": "as-is",
                                    "url": f"{host}:{port}/archive",
                                    "url_options": {"method": "GET", "data": ""},
                                },
                            },
                            {
                                "name": "job list",
                                "id": "jobs",
                                "query": {
                                    "refId": "my-query-2",
                                    "type": "json",
                                    "source": "url",
                                    "data": "",
                                    "root_selector": "",
                                    "columns": [],
                                    "filters": [],
                                    "format": "table",
                                    "url": f"{host}:{port}/jobs",
                                    "url_options": {"method": "GET", "data": ""},
                                },
                            },
                        ],
                        "refData": [],
                    },
                    "readOnly": "true",
                }
            ],
        }

    @staticmethod
    def get_dashboard_manifest(
        host="http://host.docker.internal", port=DATA_SOURCE_PORT, name=DATA_SOURCE_NAME
    ):
        return {
            "annotations": {
                "list": [
                    {
                        "builtIn": 1,
                        "datasource": {"type": "grafana", "uid": "-- Grafana --"},
                        "enable": "true",
                        "hide": "true",
                        "iconColor": "rgba(0, 211, 255, 1)",
                        "name": "Annotations & Alerts",
                        "type": "dashboard",
                    }
                ]
            },
            "editable": "true",
            "fiscalYearStartMonth": 0,
            "graphTooltip": 0,
            "id": 1,
            "links": [],
            "panels": [
                {
                    "datasource": {
                        "type": "yesoreyeram-infinity-datasource",
                        "uid": "P3E0B65AA66943F6C",
                    },
                    "fieldConfig": {
                        "defaults": {
                            "color": {"mode": "palette-classic"},
                            "custom": {
                                "axisBorderShow": "false",
                                "axisCenteredZero": "false",
                                "axisColorMode": "text",
                                "axisLabel": "",
                                "axisPlacement": "auto",
                                "barAlignment": 0,
                                "drawStyle": "line",
                                "fillOpacity": 0,
                                "gradientMode": "none",
                                "hideFrom": {
                                    "legend": "false",
                                    "tooltip": "false",
                                    "viz": "false",
                                },
                                "insertNulls": "false",
                                "lineInterpolation": "linear",
                                "lineWidth": 1,
                                "pointSize": 5,
                                "scaleDistribution": {"type": "linear"},
                                "showPoints": "auto",
                                "spanNulls": "false",
                                "stacking": {"group": "A", "mode": "none"},
                                "thresholdsStyle": {"mode": "off"},
                            },
                            "mappings": [],
                            "thresholds": {
                                "mode": "absolute",
                                "steps": [
                                    {"color": "green", "value": None},
                                    {"color": "red", "value": 80},
                                ],
                            },
                        },
                        "overrides": [],
                    },
                    "gridPos": {"h": 7, "w": 24, "x": 0, "y": 0},
                    "id": 2,
                    "options": {
                        "legend": {
                            "calcs": [],
                            "displayMode": "list",
                            "placement": "bottom",
                            "showLegend": "true",
                        },
                        "tooltip": {"mode": "single", "sort": "none"},
                    },
                    "targets": [
                        {
                            "columns": [],
                            "datasource": {
                                "type": "yesoreyeram-infinity-datasource",
                                "uid": "P3E0B65AA66943F6C",
                            },
                            "filters": [],
                            "format": "table",
                            "global_query_id": "",
                            "refId": "A",
                            "root_selector": "",
                            "source": "url",
                            "type": "json",
                            "url": f"{host}:{port}/archive",
                            "url_options": {"data": "", "method": "GET"},
                        }
                    ],
                    "title": "Task Result Status",
                    "transformations": [
                        {
                            "id": "convertFieldType",
                            "options": {
                                "conversions": [
                                    {
                                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                                        "destinationType": "time",
                                        "targetField": "timestamp",
                                    },
                                    {
                                        "destinationType": "boolean",
                                        "targetField": "result",
                                    },
                                ],
                                "fields": {},
                            },
                        }
                    ],
                    "type": "timeseries",
                },
                {
                    "datasource": {
                        "type": "yesoreyeram-infinity-datasource",
                        "uid": "P3E0B65AA66943F6C",
                    },
                    "fieldConfig": {
                        "defaults": {
                            "color": {"mode": "thresholds"},
                            "custom": {
                                "align": "auto",
                                "cellOptions": {"type": "auto"},
                                "inspect": "false",
                            },
                            "mappings": [],
                            "thresholds": {
                                "mode": "absolute",
                                "steps": [
                                    {"color": "green", "value": None},
                                    {"color": "red", "value": 80},
                                ],
                            },
                        },
                        "overrides": [],
                    },
                    "gridPos": {"h": 20, "w": 24, "x": 0, "y": 7},
                    "id": 1,
                    "options": {
                        "cellHeight": "sm",
                        "footer": {
                            "countRows": "false",
                            "fields": "",
                            "reducer": ["sum"],
                            "show": "false",
                        },
                        "showHeader": "true",
                    },
                    "pluginVersion": "10.4.0",
                    "targets": [
                        {
                            "columns": [],
                            "datasource": {
                                "type": "yesoreyeram-infinity-datasource",
                                "uid": "P3E0B65AA66943F6C",
                            },
                            "filters": [],
                            "format": "table",
                            "global_query_id": "",
                            "refId": "A",
                            "root_selector": "",
                            "source": "url",
                            "type": "json",
                            "url": f"{host}:{port}/archive",
                            "url_options": {"data": "", "method": "GET"},
                        }
                    ],
                    "title": "Panel Title",
                    "transformations": [
                        {
                            "id": "convertFieldType",
                            "options": {
                                "conversions": [
                                    {
                                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                                        "destinationType": "time",
                                        "targetField": "timestamp",
                                    },
                                    {
                                        "destinationType": "boolean",
                                        "targetField": "result",
                                    },
                                ],
                                "fields": {},
                            },
                        }
                    ],
                    "type": "table",
                },
            ],
            "schemaVersion": 39,
            "tags": [],
            "templating": {"list": []},
            "time": {"from": "now-6h", "to": "now"},
            "timepicker": {},
            "timezone": "browser",
            "title": "Task Table",
            "uid": "cdfqcy57u0gzkd",
            "version": 4,
            "weekStart": "",
        }

    def write_provisioning_files(
        self, datasources_dir: Path = None, dashboards_dir: Path = None
    ):
        # Ensure directories exist
        datasources_dir = datasources_dir or GF_DATASOURCES_DIR
        dashboards_dir = dashboards_dir or GF_DASHBOARDS_DIR

        datasources_dir.mkdir(parents=True, exist_ok=True)
        dashboards_dir.mkdir(parents=True, exist_ok=True)

        # Data source file
        data_source_path = datasources_dir / "datasource.yml"
        with open(data_source_path, "w") as file:
            yaml.dump(self.get_data_source_manifest(), file)

        # Dashboard file
        dashboard_manifest = self.get_dashboard_manifest()
        dashboard_path = dashboards_dir / f"{self.name}.json"
        with open(dashboard_path, "w") as file:
            json.dump(dashboard_manifest, file)

        # Log or handle the fact that files are written
        logger.info(f"Datasource files written: {data_source_path}")
        logger.info(f"Dashboard files written: {dashboard_path}")

    def setup_routes(self):
        @self.app.get("/jobs")
        async def get_jobs():
            # Fetch and format jobs data
            jobs_data = (
                self.timekeeper.get_jobs()
            )  # This should return a dict or list of jobs
            formatted_jobs = [
                {
                    "id": job_id,
                    "timestamp": datetime.strptime(
                        job["created"], "%Y-%m-%dT%H:%M:%S.%f"
                    ).isoformat(),
                    "task": job["task"],
                    "result": job["result"],
                    "kwargs": job["kwargs"],
                    "schedule_time": datetime.strptime(
                        job["schedule_time"], "%Y-%m-%dT%H:%M:%S.%f"
                    ),
                }
                for job_id, job in jobs_data.items()
            ]
            return formatted_jobs

        @self.app.get("/archive")
        async def get_archive():
            # Fetch and format archive data similarly
            archive_data = (
                self.timekeeper.get_archive()
            )  # This should return a dict or list of archived jobs
            formatted_archive = [
                {
                    "id": archive_id,
                    "timestamp": datetime.strptime(
                        archive["created"], "%Y-%m-%dT%H:%M:%S.%f"
                    ).isoformat(),
                    "task": archive["task"],
                    "result": archive["result"],
                    "kwargs": archive["kwargs"],
                    "schedule_time": datetime.strptime(
                        archive["schedule_time"], "%Y-%m-%dT%H:%M:%S.%f"
                    ),
                }
                for archive_id, archive in archive_data.items()
            ]
            return formatted_archive

    def start(self):
        self.logger.info("Starting DataSourceService...")
        self.server.start()
        self.logger.info("DataSourceService started.")

    def stop(self):
        self.logger.info("Stopping DataSourceService...")
        self.server.stop()
        self.logger.info("DataSourceService stopped.")



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class Task(BaseModel):
    task: str
    description: Optional[str] = None
    delay: float = 0.0  # Ensuring delay is always treated as a float
    parameters: Dict[str, Any] = Field(
        default_factory=dict
    )  # Using default_factory for parameters


class Experiment(BaseModel):
    name: Optional[str] = None
    steps: List[Task]


class ExperimentWrapper(BaseModel):
    experiment: Experiment



from typing import Callable

from PyQt6.QtWidgets import QVBoxLayout

from sonaris import factory
from sonaris.frontend.widgets.sch_scheduler import SchedulerWidget
from sonaris.frontend.widgets.templates import BasePage
from sonaris.scheduler.timekeeper import Timekeeper


class SchedulerPage(BasePage):
    PAGE_NAME = "Scheduling"

    def __init__(
        self,
        timekeeper: Timekeeper = None,
        parent=None,
        args_dict: dict = None,
        root_callback: Callable = None,
    ):
        super().__init__(
            parent=parent, args_dict=args_dict, root_callback=root_callback
        )
        self.args_dict = args_dict
        self.timekeeper = timekeeper or factory.timekeeper
        self.initUI()

    def initUI(self):
        self.main_layout = QVBoxLayout()
        self.scheduler_widget = SchedulerWidget(self.timekeeper, self.root_callback)
        # Adding the tab widget to the main layout
        self.main_layout.addWidget(self.scheduler_widget)

        self.setLayout(self.main_layout)

    def update(self):
        pass



import json
from datetime import datetime

from pydantic import BaseModel
from sqlalchemy import JSON, Boolean, Column, DateTime, String
from sqlalchemy.ext.declarative import declarative_base


class JobModel(BaseModel):
    job_id: str
    task_name: str
    schedule_time: datetime
    created: datetime
    kwargs: dict


Base = declarative_base()


class Job(Base):
    __tablename__ = "jobs"
    job_id = Column(String, primary_key=True)
    task_name = Column(String, index=True)
    schedule_time = Column(DateTime)
    created = Column(DateTime)
    kwargs = Column(JSON)
    result = Column(Boolean, default=False)
    error_info = Column(String, nullable=True)
    is_archived = Column(Boolean, default=False)  # New column to mark archived jobs

    # Convert kwargs to and from JSON automatically
    @property
    def kwargs_dict(self):
        return json.loads(self.kwargs)

    @kwargs_dict.setter
    def kwargs_dict(self, value):
        self.kwargs = json.dumps(value)



import pyvisa

# Import classes and modules from sonaris.device module as needed.
from sonaris.device.dg4202 import DG4202, DG4202DataSource, DG4202Mock
from sonaris.frontend.managers.device import DeviceManager
from sonaris.frontend.managers.state_manager import StateManager


class DG4202Manager(DeviceManager):
    device_type = DG4202
    mock_device_type = DG4202Mock

    def __init__(
        self,
        state_manager: StateManager,
        args_dict: dict,
        resource_manager: pyvisa.ResourceManager,
    ):
        super().__init__(state_manager, args_dict, resource_manager)

    def setup_data(self):
        # Will still return a valid dictionary even if self.device is None
        self.data_source = DG4202DataSource(self.device)

    def get_data(self) -> dict:
        return self.data_source.query_data()



import logging
import traceback
from datetime import datetime
from typing import Any, Callable

from apscheduler.schedulers.background import BackgroundScheduler

from sonaris.scheduler.functionmap import FunctionMap
from sonaris.utils.log import get_logger
from apscheduler.schedulers import SchedulerNotRunningError

class Worker:
    def __init__(
        self,
        function_map: dict,
        daemon: bool = False,
        logger: logging.Logger = None,
    ):
        """
        Initializes the Worker class.

        Args:
            function_map_file (Path): Path to the file containing the function map.
        """
        self.logger = logger or get_logger()
        self.scheduler = BackgroundScheduler(daemon=daemon)
        self.function_map = FunctionMap(function_map)
        self.logger.info("Function Map OK")

    def register_task(self, func: Callable, task_name: str) -> None:
        """
        Registers a function under a task name for the scheduler to recognize.

        Args:
            func (Callable): The function to be registered.
            task_name (str): The name associated with the function.
        """
        self.function_map.add_function(task_name, func)
        self.logger.info(f"Added function {task_name} {func.__name__}.")

    def __schedule_task__(
        self,
        task_name: str,
        run_time: datetime,
        job_id: str,
        _callback: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Schedules a task to be run at a specified time.

        Args:
            task_name (str): The name of the task to schedule.
            run_time (datetime): The time at which the task should run.
            job_id (str): The unique identifier of the job.
            _callback(Callable): Callback function for remove job method under Timekeeper
            *args (Any): Positional arguments to pass to the task.
            **kwargs (Any): Keyword arguments to pass to the task.
        """
        self.scheduler.add_job(
            func=self.execute_task,
            trigger="date",
            run_date=run_time,
            args=(task_name, job_id, _callback, args, kwargs),
            id=job_id,
        )
        self.logger.debug(f"Scheduled task '{task_name}' to run at {run_time}")

    def remove_scheduled_task(self, job_id: str) -> None:
        """
        Removes a scheduled task from the scheduler.

        Args:
            job_id (str): The unique identifier of the job to be removed.
        """
        try:
            self.scheduler.remove_job(job_id)
            self.logger.info(
                f"Successfully removed scheduled task with job_id: {job_id}."
            )
        except Exception as e:
            self.logger.error(
                f"Error removing scheduled task with job_id: {job_id}. Error: {e}"
            )

    def start_worker(self) -> None:
        """
        Starts the worker.
        """
        self.scheduler.start()
        self.logger.debug("APScheduler worker started.")

    def stop_worker(self) -> None:
        """
        Stops the worker.
        """
        try:
            self.scheduler.shutdown()
            self.logger.info("APScheduler worker stopped.")
        except SchedulerNotRunningError as e:
            self.logger.info("APScheduler is already stopped.")
        except Exception as e:
            self.logger.error(f"Error stopping APScheduler: {e}.")

    def execute_task(
        self,
        task_name: str,
        job_id: str = None,
        _callback: Callable = None,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Executes a registered task and removes it from the schedule after execution.

        Args:
            task_name (str): The name of the task to execute.
            job_id (str, optional): The unique identifier of the job.
            _callback (Callable, optional): The callback function to remove the job.
            *args (Any): Positional arguments to pass to the task.
            **kwargs (Any): Keyword arguments to pass to the task.

        Returns:
            Any: The result of the task function, if any.
        """
        task_func = self.function_map.get_function(task_name)
        result = False  # Initialize false
        error_info = None  # Initialize a variable to store exception info
        if task_func:
            self.logger.info(f"Executing task '{task_name}'(id:{job_id}).")
            try:
                self.function_map.parse_and_call(task_func, *args, **kwargs)
                result = True
                self.logger.info(
                    f"Task '{task_name}' successfully executed."
                )  # Changed from error to info
            except Exception as e:
                error_info = traceback.format_exc()  # Capture and format the traceback
                self.logger.error(f"Task '{task_name}' failed to execute. Error: {e}")
                self.logger.error(f"{error_info}")
        else:
            self.logger.error(f"Task '{task_name}' is not registered.")
        if _callback and job_id:
            _callback(
                job_id, result, error_info
            )  # Modified to pass the traceback information
        return result



from unittest.mock import MagicMock

import numpy as np

from sonaris.device.data import DataSource
from sonaris.device.device import Device, MockDevice
from sonaris.device.interface import Interface


class EDUX1002A(Device):
    IDN_STRING = "EDU-X 1002A"

    """Keysight EDUX1002A hardware driver/wrapper."""

    def __init__(self, interface: Interface, timeout=20000):
        super().__init__(interface)
        self.interface.inst.timeout = timeout

    def initialize(self):
        """Reset and clear the oscilloscope to default settings."""
        self.interface.write("*RST")
        self.interface.write("*CLS")

    def autoscale(self):
        """Use Autoscale for automatic oscilloscope setup."""
        self.interface.write(":AUToscale")

    def set_acquisition_complete(self, percentage: int):
        """
        Set the percent completion for the acquisition process.

        Parameters:
        - percentage (int): Percentage of completion. An integer between 0 and 100.
        """
        if not 0 <= percentage <= 100:
            raise ValueError("Percentage should be an integer between 0 and 100.")

        self.interface.write(f":ACQuire:COMPlete {percentage}")

    def set_waveform_source(self, channel: int = 1):
        """
        Set the waveform source.

        Parameters:
        - channel (int): The channel to set as the waveform source.
        """
        self.interface.write(f":WAVeform:SOURce CHANnel{channel}")

    def set_waveform_format(self, format: str):
        """
        Set the waveform data format.

        Parameters:
        - format (str): The format for the waveform data. Options include "BYTE", "WORD", and "ASCII".
        """
        valid_formats = ["BYTE", "WORD", "ASCII"]
        if format not in valid_formats:
            raise ValueError(f"Invalid format. Choose one of {valid_formats}.")

        self.interface.write(f":WAVeform:FORMat {format}")

    def set_waveform_points(self, points: int):
        """
        Set the number of waveform points to retrieve.

        Parameters:
        - points (int): The number of points to retrieve.
        """
        self.interface.write(f":WAVeform:POINts {points}")

    def get_waveform_data_raw(self):
        """
        Get the raw waveform data from the oscilloscope.

        Returns:
        - str: The raw waveform data.
        """
        return self.interface.read(":WAVeform:DATA?")

    def set_trigger_mode(self, mode="EDGE"):
        """Set the trigger mode of the oscilloscope."""
        self.interface.write(f":TRIGger:MODE {mode}")

    def digitize(self, channel=0):
        """Capture data using the :DIGitize command."""
        if channel == 0:
            self.interface.write(":DIGitize")
        else:
            self.interface.write(f":DIGitize CHANnel{channel}")

    def query_oscilloscope(self, query):
        """Read query responses from the oscilloscope."""
        return self.interface.read(query)

    def check_instrument_status(self):
        """Check and print the instrument's status."""
        status = self.interface.read(":SYSTem:ERRor?")
        print(f"Oscilloscope Status: {status}")
        return status

    def set_acquisition_mode(self, mode="RTIMe"):
        """
        Set the acquisition mode of the oscilloscope.

        Parameters:
        - mode (str): The acquisition mode to set. Options are "RTIMe" for real-time mode
                      and "SEGMented" for segmented mode. Default is "RTIMe".
        """
        if mode not in ["RTIMe", "SEGMented"]:
            raise ValueError("Invalid acquisition mode. Choose 'RTIMe' or 'SEGMented'.")

        self.interface.write(f":ACQuire:MODE {mode}")

    def is_real_time_mode(self):
        """Check if the oscilloscope is in real-time mode."""
        current_mode = self.interface.read(":ACQuire:MODE?")
        return current_mode.strip() == "RTIMe"

    def setup_waveform_readout(self, channel: int = 1, mode="ASCII"):
        """Setup the oscilloscope for waveform readout."""
        # self.interface.write(f"CHANNEL{channel}:DISPLAY ON")
        # self.interface.write(f"DATA:SOURCE CHANNEL{channel}")
        # self.set_waveform_format(mode)
        self.set_waveform_source(channel)

    def get_waveform_preamble(self):
        """Retrieve the waveform preamble which provides data on the waveform format."""
        preamble_str = self.interface.read(":WAVeform:PREamble?")
        preamble_values = preamble_str.split(",")
        # print(preamble_str)
        # Convert the values according to the documentation
        preamble = [
            int(preamble_values[0]),  # format
            int(preamble_values[1]),  # type
            int(preamble_values[2]),  # points
            int(preamble_values[3]),  # count
            float(preamble_values[4]),  # xincrement
            float(preamble_values[5]),  # xorigin
            int(preamble_values[6]),  # xreference
            float(preamble_values[7]),  # yincrement
            float(preamble_values[8]),  # yorigin
            int(preamble_values[9]),  # yreference
        ]

        return preamble

    def is_connection_alive(self) -> bool:
        """
        Checks if the connection to the EDUX1002A device is alive.

        Returns:
            bool: True if the connection is alive, False otherwise.
        """
        try:
            response = self.interface.read("*IDN?")
            return bool(response)  # True if there is a response, False otherwise
        except:
            return False

    def display_preamble_details(self, preamble):
        """Displays the preamble details in a human-readable format."""

        # Expected names and their explanations based on documentation
        names = [
            "Format",
            "Type",
            "Points",
            "Count",
            "X Increment",
            "X Origin",
            "X Reference",
            "Y Increment",
            "Y Origin",
            "Y Reference",
        ]

        format_explanations = {0: "BYTE format", 1: "WORD format", 4: "ASCII format"}

        type_explanations = {
            0: "NORMal type",
            1: "PEAK detect type",
            2: "AVERage type",
        }

        # Extracting the values
        details = [
            format_explanations.get(preamble[0], f"Unknown Value {preamble[0]}"),
            type_explanations.get(preamble[1], f"Unknown Value {preamble[1]}"),
        ] + preamble[2:]

        # Displaying the details
        result_str = "Preamble Details:\n"
        for name, detail in zip(names, details):
            result_str += f"{name}: {detail}\n"

        print(result_str)

    def get_waveform_data(self, channel: int = 1):
        """Get the waveform data from the oscilloscope."""
        self.setup_waveform_readout(channel)
        waveform_data = self.get_waveform_data_raw()
        preamble = self.get_waveform_preamble()

        # Check for header
        if waveform_data[0] == "#":
            num_digits = int(waveform_data[1])
            num_data_points = int(waveform_data[2 : 2 + num_digits])

            # Extract the actual data without the header
            waveform_data = waveform_data[2 + num_digits :]

        format_type = preamble[0]
        print(self.display_preamble_details(preamble))
        if format_type == 4:  # ASCII
            return preamble, np.array([float(val) for val in waveform_data.split(",")])
        elif format_type == 0:  # BYTE
            return preamble, np.frombuffer(
                waveform_data.encode("latin-1"), dtype=np.int8
            )
        elif format_type == 1:  # WORD
            return preamble, np.frombuffer(
                waveform_data.encode("latin-1"), dtype=np.int16
            )
        else:
            raise ValueError("Unknown waveform format.")

    def get_waveform(self, channel: int = 1):
        """Public method to setup, retrieve, and process waveform data."""
        self.digitize(channel)
        preamble, waveform_data = self.get_waveform_data(channel)

        # Extract information from preamble
        x_increment = preamble[4]
        x_origin = preamble[5]
        y_increment = preamble[7]
        y_origin = preamble[8]

        # Convert data to actual voltage and time values
        time = np.arange(len(waveform_data)) * x_increment + x_origin
        voltage = waveform_data * y_increment + y_origin

        return time, voltage

    def set_timeout(self, timeout):
        self.interface.inst.timeout = timeout

    def set_acquisition_type(self, acq_type="NORMal"):
        """
        Set the data acquisition type of the oscilloscope.

        Parameters:
        - acq_type (str): The acquisition type to set. Options are "NORMal", "AVERage",
                          "HRESolution", and "PEAK". Default is "NORMal".
        """
        valid_types = ["NORMal", "AVERage", "HRESolution", "PEAK"]
        if acq_type not in valid_types:
            raise ValueError(f"Invalid acquisition type. Choose one of {valid_types}.")

        self.interface.write(f":ACQuire:TYPE {acq_type}")

    def set_waveform_return_type(self, ret_type="NORMal"):
        """
        Set the data acquisition type of the oscilloscope.

        Parameters:
        - ret_type (str): The acquisition type to set. Options are "NORMal", "AVERage",
                          "HRESolution", and "PEAK". Default is "NORMal".
        """
        valid_types = ["NORMal", "PEAK", "AVERage", "HREsolution"]
        if ret_type not in valid_types:
            raise ValueError(f"Invalid acquisition type. Choose one of {valid_types}.")

        self.interface.write(f":WAVeform:TYPE {ret_type}")

    def get_acquisition_type(self):
        """
        Query the oscilloscope for the current acquisition type.

        Returns:
        - str: The current acquisition type. One of "NORM", "AVER", "HRES", or "PEAK".
        """
        return self.interface.read(":ACQuire:TYPE?").strip()

    def set_acquisition_count(self, count=1):
        """
        Set the acquisition count. Relevant only for "AVERage" acquisition type.

        Parameters:
        - count (int): The number of averages. An integer from 1 to 65536.
        """
        if not 1 <= count <= 65536:
            raise ValueError(
                "Acquisition count should be an integer between 1 and 65536."
            )

        self.interface.write(f":ACQuire:COUNt {count}")


class EDUX1002ADataSource(DataSource):
    def __init__(self, source: EDUX1002A, channel: int = 1):
        super().__init__(source)
        self.channel = channel

    def query_data(self):
        try:
            time, voltage = self.source.get_waveform(self.channel)
            return voltage
        except Exception as e:
            raise RuntimeError(f"Error querying EDUX1002A:{e}")


class EDUX1002AMockInterface(Interface):
    def __init__(self):
        # Create a MagicMock instance to simulate a pyvisa.Resource
        mock_resource = MagicMock()
        mock_resource.timeout = None  # Set default value for timeout attribute
        super().__init__(mock_resource)
        self.state = {
            ":WAVeform:DATA?": "Mocked Waveform Data",
            ":SYSTem:ERRor?": "No error",
        }
        self.debug = True

    def write(self, command: str) -> None:
        # Use the mocked write method
        self.inst.write(command)

    def read(self, command: str) -> str:
        # Simulate reading a response from the device
        return self.state.get(command, self.inst.query(command))


class EDUX1002AMock(MockDevice, EDUX1002A):
    def __init__(self, timeout=20000):
        interface = EDUX1002AMockInterface()
        super().__init__(interface=interface)
        EDUX1002A.__init__(self, interface=interface, timeout=timeout)
        self.blocked_methods = {
            "initialize",
            "autoscale",
            "set_waveform_source",
            "set_waveform_format",
            "get_waveform_data_raw",
            "set_trigger_mode",
            "digitize",
            "query_oscilloscope",
            "check_instrument_status",
            "set_acquisition_mode",
            "is_real_time_mode",
            "get_waveform_preamble",
            "get_waveform_data",
            "get_waveform",
            "set_timeout",
            "set_acquisition_type",
            "set_waveform_return_type",
            "get_acquisition_type",
            "set_acquisition_count",
        }

    # Implementing the mocked methods
    def initialize(self):
        print("Oscilloscope reset to default settings.")

    def autoscale(self):
        print("Autoscale completed.")

    def set_waveform_source(self, channel: int = 1):
        print(f"Waveform source set to Channel {channel}.")

    def set_waveform_format(self, format: str):
        print(f"Waveform format set to {format}.")

    def get_waveform_data_raw(self):
        return "Mocked Waveform Data"

    def get_waveform_data(self, channel: int = 1):
        # Return simulated waveform data
        return np.random.rand(100)  # Simulating 100 data points

    def get_waveform(self, channel: int = 1):
        time = np.arange(100)  # Simulating time data
        voltage = self.get_waveform_data(channel)
        return time, voltage



from threading import Thread
from typing import List, Optional, Tuple, Union

from flask import Flask, jsonify, request
from flask.wrappers import Response
from werkzeug.serving import make_server

from sonaris.device.dg4202 import DG4202, DG4202Mock


class DG4202APIServer:

    def __init__(self, dg4202: DG4202, server_port: int = 5000) -> None:
        """
        Create a new DG4202API instance.

        Args:
            dg4202 (DG4202): A DG4202 instance.
            server_port (int): Default port is 5000.
        """
        self.dg4202 = dg4202
        self.http_server = None
        self.app = Flask(__name__)
        self.server_port = server_port
        self.server_thread = None
        self.setup_routes()

    def shutdown(self) -> None:
        """
        Shutdown the Flask application server.
        """
        if self.http_server:
            self.http_server.shutdown()

    def setup_routes(self) -> None:
        """
        Setup Flask routes for the API.
        """

        @self.app.route("/api/command", methods=["POST"])
        def send_command() -> Union[Response, Tuple[Response, int]]:
            """
            Flask route for sending a command to the DG4202.

            Returns:
                Response: a Flask response.
            """
            command = request.json.get("command")
            if isinstance(self.dg4202, DG4202Mock):
                if self.dg4202.killed:
                    return jsonify({"error": f"{command} failed."}), 400

            if command is None:
                return jsonify({"error": f"{command} failed."}), 400
            self.dg4202.interface.write(command)
            return jsonify({"status": f"{command} sent"}), 200

        @self.app.route("/api/simulate_kill", methods=["POST"])
        def simulate_kill() -> Union[Response, Tuple[Response, int]]:
            """
            Flask route for sending a command to the DG4202.

            Returns:
                Response: a Flask response.
            """
            if isinstance(self.dg4202, DG4202Mock):
                kill = request.json.get("kill")
                if kill is not None:
                    if kill == "true":
                        self.dg4202.simulate_kill(True)
                        return jsonify({"status": f"{kill} sent"}), 200
                    elif kill == "false":
                        self.dg4202.simulate_kill(False)
                        return jsonify({"status": f"{kill} sent"}), 200
                return jsonify({"error": f"{kill} failed."}), 400
            else:
                return jsonify({"error": "interface is not a hardware mock"}), 400

        @self.app.route("/api/state", methods=["GET"])
        def get_state() -> Response:
            """
            Flask route for retrieving the state of the DG4202.

            Returns:
                Response: a Flask response.
            """
            state = request.args.get("state")
            if isinstance(self.dg4202, DG4202Mock):
                if self.dg4202.killed:
                    return jsonify({"error": f"{state} failed."}), 400

            if state is None:
                return jsonify({"error": "state parameter is missing."}), 422

            # Assuming self.dg4202_interface.read() accepts the state parameter
            state_value = self.dg4202.interface.read(state)

            return jsonify({"state": state_value}), 200

        @self.app.route("/api/stop", methods=["POST"])
        def stop_server():
            """
            Flask route to stop the server.
            """
            self.stop()
            return jsonify({"status": "Server shutting down..."}), 200

    def run(self, port: int = None) -> None:
        """
        Start the Flask application in a new thread.

        Args:
            port (int, optional): The port number to listen on. Defaults to 5000.
        """
        self.server_thread = Thread(target=self._run, args=(port,))
        self.server_thread.start()

    def _run(self, port: int) -> None:
        """
        Start the Flask application.

        Args:
            port (int): The port number to listen on.
        """
        if port is None:
            port = self.server_port

        self.http_server = make_server("0.0.0.0", port, self.app)
        self.http_server.serve_forever()



from pathlib import Path
from typing import Optional

from docker import DockerClient
from docker.models.containers import Container

from sonaris.defaults import (
    APP_NAME,
    GF_DASHBOARDS_DIR,
    GF_DATASOURCES_DIR,
    GF_PORT,
    GF_SECURITY_ADMIN_PASSWORD,
    SONARIS_NETWORK_NAME,
)
from sonaris.services.container_service import ContainerService
from sonaris.utils.log import get_logger

logger = get_logger()

from sonaris.defaults import GF_INSTALL_PLUGINS, GF_PORT, GF_SECURITY_ADMIN_PASSWORD


class GrafanaService(ContainerService):
    def __init__(
        self,
        client: DockerClient = None,
        port: int = None,
        image: str = None,
        datasources_dir: Path = None,
        dashboards_dir: Path = None,
        network_name: str = None,
        plugins: str = None,
    ):
        super().__init__(
            client=client,
            network_name=network_name if network_name else SONARIS_NETWORK_NAME,
            service_name="grafana",
            image=image or "grafana/grafana",
        )
        self.plugins = plugins or GF_INSTALL_PLUGINS
        self.port = port or GF_PORT
        self.datasources_dir = datasources_dir or GF_DATASOURCES_DIR
        self.dashboards_dir = dashboards_dir or GF_DASHBOARDS_DIR

    def create_container(self, image: str = "grafana/grafana") -> Optional[str]:
        if self.client is None:
            logger.error("Docker client is unavailable. Cannot start Grafana service.")
            return None
        try:
            container: Container = self.client.containers.run(
                image,
                ports={"3000/tcp": self.port},
                environment={
                    "GF_SECURITY_ADMIN_PASSWORD": f"{GF_SECURITY_ADMIN_PASSWORD}",
                    "GF_INSTALL_PLUGINS": self.plugins,
                },
                volumes={
                    str(self.datasources_dir): {
                        "bind": "/etc/grafana/provisioning/datasources",
                        "mode": "rw",
                    },
                    str(self.dashboards_dir): {
                        "bind": "/etc/grafana/provisioning/dashboards",
                        "mode": "rw",
                    },
                },
                labels={f"grafana": self._container_label},
                detach=True,
            )
            if self.network:
                self.network.connect(container)
            logger.info(
                f"Grafana container created and started. Accessible on http://localhost:{self.port}."
            )
            return container.id
        except Exception as e:
            logger.error(f"Failed to create Grafana container: {e}")
            return None



from pathlib import Path
from typing import Callable

from PyQt6.QtWidgets import QVBoxLayout

from sonaris.frontend.widgets.mon_device import (  # Assuming this is the path
    DeviceMonitorWidget,
)
from sonaris.frontend.widgets.templates import BasePage


class MonitorPage(BasePage):
    PAGE_NAME = "Monitor"

    def __init__(
        self,
        device_managers: dict,
        parent=None,
        args_dict: dict = None,
        monitor_logs: Path = None,
        root_callback: Callable = None,
    ):
        super().__init__(
            parent=parent, args_dict=args_dict, root_callback=root_callback
        )
        self.device_managers = device_managers
        self.monitor_logs = monitor_logs
        self.initUI()

    def initUI(self):
        self.main_layout = QVBoxLayout()
        self.device_monitor_widget = DeviceMonitorWidget(
            self.device_managers, monitor_logs=self.monitor_logs
        )
        self.main_layout.addWidget(self.device_monitor_widget)
        self.setLayout(self.main_layout)

    def update(self):
        # Implement any necessary updates when the page is loaded or refreshed
        pass



import json
import logging
import os
from logging.handlers import RotatingFileHandler
from pathlib import Path

from colorlog import ColoredFormatter

from sonaris.defaults import LOG_DIR

# Global logger variable
logger = None


def load_json_with_backup(path: Path):
    """
    Attempts to load a JSON file from the given path. If the file is corrupt,
    it creates a numbered backup and returns an empty dictionary.
    """
    if path.exists():
        try:
            with open(path, "r") as file:
                return json.load(file)
        except json.JSONDecodeError as e:
            logger.info(f"Corrupt JSON file detected: {e}. Creating a numbered backup.")
            backup_path = create_numbered_backup(path)
            path.rename(backup_path)
        except Exception as e:
            logger.info(f"Unexpected error loading JSON file: {e}.")
    return {}


def save_json(data, path: Path):
    """
    Saves the given data as a JSON file to the specified path.
    """
    try:
        with open(path, "w") as file:
            json.dump(data, file, indent=4)
    except Exception as e:
        logger.info(f"Unexpected error loading JSON file: {e}.")


def create_numbered_backup(original_path: Path):
    """
    Creates a numbered backup for the given path, ensuring no existing backup is overwritten.
    """
    backup_base = original_path.with_suffix(".bak")
    counter = 1
    new_backup = Path(f"{backup_base}_{counter}")
    while new_backup.exists():
        counter += 1
        new_backup = Path(f"{backup_base}_{counter}")
    return new_backup


def init_logging(logger_name: str = None):
    global logger
    if logger is None:
        logger_name = logger_name or "sonaris"
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)

        logs_path = LOG_DIR
        if not logs_path.is_dir():
            logs_path.mkdir(parents=True, exist_ok=True)
        log_file_path = logs_path / f"{logger_name}.log"
        logger.propagate = False
        # Handler for writing logs to a file
        file_handler = RotatingFileHandler(
            filename=str(log_file_path), maxBytes=10000000, backupCount=5
        )
        file_handler.setLevel(logging.DEBUG)

        # Handler for printing logs to the console
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        colored_formatter = ColoredFormatter(
            "%(log_color)s%(asctime)s - %(name)s - %(levelname)s - %(module)s.%(funcName)s:\033[97m%(lineno)d\033[0m - %(message)s",
            datefmt=None,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
            secondary_log_colors={},
            style="%",
        )

        # Adjusted formatter to include module names
        # formatter = logging.Formatter(
        #     "%(asctime)s - %(name)s - %(levelname)s - %(module)s.%(funcName)s:%(lineno)d - %(message)s"
        # )

        file_handler.setFormatter(colored_formatter)
        console_handler.setFormatter(colored_formatter)

        # Adding both handlers to the logger
        logger.addHandler(file_handler)
        logger.addHandler(console_handler)


def get_logger(module_name=None):
    if logger is None:
        init_logging()
    if module_name:
        return logging.getLogger(f"{os.getenv('LOGGER_NAME', 'logs')}.{module_name}")
    return logger



import abc
import logging
import time
from datetime import timedelta
from typing import Type, Union

import pyvisa

# Import classes and modules from sonaris.device module as needed.
from sonaris.device.data import DataSource
from sonaris.device.device import Device, DeviceDetector, MockDevice
from sonaris.frontend.managers.state_manager import StateManager
from sonaris.utils.log import get_logger

logger = get_logger()


class DeviceManager(abc.ABC):
    device_type: Type[Device] = Device
    mock_device_type: Type[MockDevice] = MockDevice

    def __init__(
        self,
        state_manager: StateManager,
        args_dict: dict,
        resource_manager: pyvisa.ResourceManager,
    ):
        self.state_manager = state_manager
        self.args_dict = args_dict
        self.resource_manager = resource_manager
        self.mock_device = self.mock_device_type()
        self.setup_device()
        self.setup_data()

    def setup_device(self) -> None:
        if self.args_dict.get("hardware_mock", False):
            self.fetch_mock_hardware()
        else:
            self.fetch_hardware()

    def fetch_hardware(self) -> None:
        """Fetch and update the device driver (hardware, not simulated!)"""
        self.device = DeviceDetector(
            resource_manager=self.resource_manager, device_type=self.device_type
        ).detect_device()

    def fetch_mock_hardware(self) -> None:
        """Sets the internal pointer of device to the mock device.
        Sets it to None if the mock device is simulated to be killed.
        """
        self.device = self.mock_device if not self.mock_device.killed else None

    def setup_data(self):
        # Will still return a valid dictionary even if self.device is None
        self.data_source = DataSource(self.device)
        raise NotImplementedError("Must be implemented in subclasses.")

    def update_last_alive_state(self, last_alive_key: str, alive: bool) -> None:
        """
        Updates the 'last_alive' timestamp in the state for the device.

        Parameters:
            last_alive_key (str): The key in the state dict corresponding to the device's last alive timestamp.
            alive (bool): Indicates whether the device is considered alive or not.
        """
        state = self.state_manager.read_state()
        if alive:
            state[last_alive_key] = time.time()
        else:
            state[last_alive_key] = None
        self.state_manager.write_state(state)

    def get_device(self) -> Union[Device, MockDevice, None]:
        last_alive_key = f"{self.device_type.IDN_STRING}_last_alive"

        # Decide if we should use a mock device or try to fetch a real device.
        if self.args_dict.get("hardware_mock", False):
            self.fetch_mock_hardware()
        else:
            self.fetch_hardware()  # This attempts to set `self.device` to a real device.
        # Determine if the device is considered 'alive'.
        device_alive = self.device is not None and (
            not isinstance(self.device, MockDevice) or not self.device.killed
        )
        self.update_last_alive_state(last_alive_key, device_alive)
        self.setup_data()
        return self.device

    def set_mock_state(self, state: bool) -> None:
        self.mock_device.killed = state

    def call_device_method(self, method_name: str, *args, **kwargs):
        """
        Generic method to call a method on the managed device.

        :param method_name: The name of the method to be called on the device.
        :param args: Positional arguments to pass to the device method.
        :param kwargs: Keyword arguments to pass to the device method.
        :return: The result of the device method call.
        """
        self.device = self.get_device()  # Ensure we have the current device instance
        if self.device is not None:
            try:
                method = getattr(self.device, method_name)
                if callable(method):
                    return method(*args, **kwargs)
                else:
                    raise AttributeError(
                        f"{method_name} is not a method of {self.device.IDN_STRING}"
                    )
            except AttributeError as e:
                logging.error(
                    f"Method {method_name} not found on device {self.device.IDN_STRING}: {e}"
                )
                return None
        else:
            logging.error(f"No device instance available for {self.device.IDN_STRING}")
            return None

    def is_device_alive(self) -> bool:
        try:
            if self.args_dict["hardware_mock"]:
                return not self.device.killed
            idn = self.device.interface.read("*IDN?")
            return self.device.IDN_STRING in idn
        except Exception as e:
            return False

    def write_device_state(self) -> None:
        alive = self.is_device_alive()
        if alive:
            self.state_manager.update_device_last_alive(
                self.device.IDN_STRING, time.time()
            )
        else:
            self.state_manager.update_device_last_alive(self.device.IDN_STRING, None)

    def get_device_uptime(self) -> str:
        last_alive = self.state_manager.get_device_last_alive(self.device.IDN_STRING)
        if last_alive:
            uptime_seconds = time.time() - last_alive
            return str(timedelta(seconds=int(uptime_seconds)))
        else:
            return "N/A"



import hashlib
import json
import logging
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Callable, Dict

from sonaris.defaults import DEFAULT_DATADIR
from sonaris.scheduler.worker import Worker
from sonaris.utils.log import create_numbered_backup, get_logger


class Timekeeper:
    def __init__(
        self,
        persistence_file: Path,
        worker_instance: Worker,
        logger: logging.Logger = None,
        user_callback: Callable = None,
        archive: Path = None,
    ):
        """
        Initializes the Timekeeper class, responsible for managing and scheduling jobs.

        Args:
            persistence_file (Path): Path to the file used for persisting job data.
            worker_instance (Worker): An instance of the Worker class to execute scheduled tasks.
        """
        self.logger = logger or get_logger()
        self.persistence_file = persistence_file
        self.worker = worker_instance
        self.jobs = self.load_jobs()
        self.archive = (
            archive
            or Path(os.getenv("DATA"), "archive.json")
            or DEFAULT_DATADIR / "archive.json"
        )
        self.reload_function_map()
        self.__reschedule_jobs__()
        self.user_callback = user_callback
    def get_archive(self) -> Dict[str, Any]:
        return json.loads(self.archive.read_text())
    def set_callback(self, user_callback: Callable) -> None:
        self.user_callback = user_callback

    def cancel_job(self, job_id: str) -> None:
        """Removes job from worker and erases entry

        Args:
            job_id (str): _description_
        """
        self.worker.remove_scheduled_task(job_id)
        self.remove_job(job_id)

    def clear_archive(self):
        # Clear the JSON file
        try:
            with self.archive.open("w") as file:
                json.dump({}, file)
        except Exception as e:
            self.logger.error(f"Error clearing finished jobs: {e}")

    def load_jobs(self) -> Dict[str, Any]:
        """
        Loads the jobs from the persistence file.

        Returns:
            Dict[str, Any]: A dictionary of jobs indexed by their IDs.
        """
        try:
            with open(self.persistence_file, "r") as file:
                return json.load(file)
        except (FileNotFoundError, json.JSONDecodeError):
            return {}

    def save_jobs(self) -> None:
        """
        Saves the current jobs to the persistence file.
        """
        with open(self.persistence_file, "w") as file:
            json.dump(self.jobs, file, indent=4)

    def compute_hash(
        self, task_name: str, schedule_time: datetime, *args, **kwargs
    ) -> str:
        """
        Computes a hash for the given task and arguments.

        Args:
            task_name (str): The name of the task.
            *args: Positional arguments for the task.
            **kwargs: Keyword arguments for the task.

        Returns:
            str: A hash string representing the task and arguments.
        """
        args = args or ""
        kwargs = kwargs or ""
        return hashlib.sha256(
            str(f"{task_name}{schedule_time.isoformat()}{args}{kwargs}").encode()
        ).hexdigest()[:12]

    def add_job(self, task_name: str, schedule_time: datetime, **kwargs) -> str:
        """
        Adds a new job to the schedule.

        Args:
            task_name (str): The name of the task to schedule.
            schedule_time (datetime): The time at which the task should be executed.
            **kwargs: Keyword arguments to pass to the task.

        Returns:
            str: The ID of the scheduled job.
        """
        job_id = self.compute_hash(task_name, schedule_time, kwargs)
        self.jobs[job_id] = {
            "task": task_name,
            "created": datetime.now().isoformat(),
            "schedule_time": schedule_time.isoformat(),
            **kwargs,
        }
        self.save_jobs()
        self.logger.info(
            f"Received job {job_id} with task {task_name} to run at {schedule_time}"
        )
        self.schedule_job_to_worker(job_id)
        return job_id

    def schedule_job_to_worker(self, job_id: str) -> None:
        """
        Schedules a job to be executed by the worker.

        Args:
            job_id (str): The ID of the job to schedule.
        """
        job_info = self.jobs[job_id]
        schedule_time = datetime.fromisoformat(job_info["schedule_time"])
        self.worker.__schedule_task__(
            job_info["task"],
            schedule_time,
            job_id,
            self.callback,
            **job_info["kwargs"],
        )

    def reload_function_map(self) -> None:
        """
        Reloads the function map from the Worker instance.
        """
        try:
            for (
                func_identifier,
                func_data,
            ) in self.worker.function_map.function_map.items():
                func = self.worker.function_map.deserialize_func(func_data)
                self.worker.register_task(func, func_identifier)
            self.logger.debug("Function map reloaded.")
        except Exception as e:
            self.logger.error(f"Failed to reload function map: {e}")
            backup_file = create_numbered_backup(self.persistence_file)
            self.logger.info(
                f"Created numbered backup of the function map: {backup_file}"
            )

    def __reschedule_jobs__(self) -> None:
        """
        Reschedules all jobs that are due to run.
        """
        self.logger.debug(f"Found {len(self.jobs)} scheduled.")
        now = datetime.now()
        self.prune()
        for job_id, job_info in self.jobs.items():
            schedule_time = datetime.fromisoformat(job_info["schedule_time"])
            if schedule_time < now:
                schedule_time = now + timedelta(seconds=10)
                self.logger.info(f"Rescheduling job {job_id} to run at {schedule_time}")
            self.worker.__schedule_task__(
                job_info["task"],
                schedule_time,
                job_id,
                self.callback,
                **job_info["kwargs"],
            )

    def archive_job(self, job_id: str, job_info: Dict[str, Any]) -> None:
        """
        Archives a completed job.

        Args:
            job_id (str): The ID of the completed job.
            job_info (Dict[str, Any]): The details of the completed job.
        """
        try:
            if not self.archive.exists():
                # Create the archive file if it doesn't exist
                with open(self.archive, "w") as file:
                    json.dump({}, file)

            with open(self.archive, "r+") as file:
                archived_jobs = json.load(file)
                archived_jobs[job_id] = job_info
                file.seek(0)
                json.dump(archived_jobs, file, indent=4)

            self.logger.info(f"Job {job_id} archived.")
        except Exception as e:
            self.logger.error(f"Failed to archive job {job_id}: {e}")

    def callback(self, job_id: str, result: bool, error_info: str = None) -> None:
        """
        Callback method that is called when a job is completed.

        Args:
            job_id (str): The unique identifier of the job.
            result (bool): The result of the job execution, True if successful, False otherwise.
            error_info (str, optional): The traceback or error information if the job failed.
        """
        # Retrieve the job information, if not found, use an empty dictionary
        job_info = self.jobs.get(job_id, {})

        # Update the job_info dictionary with the result of the execution
        job_info["result"] = result

        # If there is error information (job execution failed), add it to the job_info
        if error_info:
            job_info["error_info"] = error_info

        # Perform job archival and removal
        self.archive_job(
            job_id, job_info
        )  # Assuming you want to archive the job with updated info
        self.remove_job(job_id)

        # Call the user-defined callback if it exists
        if self.user_callback is not None:
            # You might want to pass additional arguments or the updated job_info to your user_callback
            self.user_callback()

    def remove_job(self, job_id: str) -> None:
        """Removes from internal entry, not on worker node!

        Args:
            job_id (str): _description_
        """
        self.jobs.pop(job_id)
        self.save_jobs()
        self.logger.info(f"Job {job_id} removed.")

    def prune(self) -> None:
        """
        Removes jobs that are no longer valid or have passed their schedule time.
        """
        now = datetime.now()
        jobs_to_remove = [
            job_id
            for job_id, job_info in self.jobs.items()
            if datetime.fromisoformat(job_info["schedule_time"]) < now
        ]
        for job_id in jobs_to_remove:
            del self.jobs[job_id]
            self.logger.info(f"Pruned job {job_id}")
        self.save_jobs()

    def get_jobs(self) -> Dict[str, Any]:
        """
        Returns the currently scheduled jobs.

        Returns:
            Dict[str, Any]: A dictionary of the scheduled jobs.
        """
        return self.jobs
